from django.conf import settings
from django.contrib.auth import authenticate, get_user_model
from django.contrib.auth.models import AnonymousUser, Group, Permission, User
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.http import Http404, HttpResponse
from django.shortcuts import resolve_url
from django.test import RequestFactory, TestCase, override_settings
from django.views import View

from djem.auth import (
    ObjectPermissionsBackend,
    PermissionRequiredMixin,
    permission_required,
)

from .models import CustomUser, OLPTest, UniversalOLPTest, UserLogTest

_backends = [
    'django.contrib.auth.backends.ModelBackend',
    'djem.auth.ObjectPermissionsBackend'
]


def _test_view(request, obj=None):
    
    return HttpResponse('success')


class _TestView(PermissionRequiredMixin, View):
    
    def get(self, *args, **kwargs):
        
        return HttpResponse('success')


@override_settings(AUTH_USER_MODEL='djemtest.CustomUser', AUTHENTICATION_BACKENDS=_backends)
class OLPMixinTestCase(TestCase):
    
    def setUp(self):
        
        self.user = CustomUser.objects.create_user('test.user')
    
    def test_clear_perm_cache__no_checks(self):
        """
        Test the clear_perm_cache() method before any permissions have been
        checked. It should have no effect.
        """
        
        user = self.user
        
        # OLPMixin defines the cache attribute by default, but it should be empty
        self.assertEqual(user._olp_cache, {})
        
        # Django's MLP cache attributes should not exist
        with self.assertRaises(AttributeError):
            user._user_perm_cache  # noqa: B018
        
        with self.assertRaises(AttributeError):
            user._group_perm_cache  # noqa: B018
        
        with self.assertRaises(AttributeError):
            user._perm_cache  # noqa: B018
        
        # Clear the cache
        user.clear_perm_cache()
        
        self.assertEqual(user._olp_cache, {})
        
        with self.assertRaises(AttributeError):
            user._user_perm_cache  # noqa: B018
        
        with self.assertRaises(AttributeError):
            user._group_perm_cache  # noqa: B018
        
        with self.assertRaises(AttributeError):
            user._perm_cache  # noqa: B018
    
    def test_clear_perm_cache__mlp(self):
        """
        Test the clear_perm_cache() method after model-level permissions have
        been checked. It should remove the attributes used by Django for
        various MLP caches.
        """
        
        user = self.user
        
        # Grant the user some permissions - both directly and via a group
        permissions = Permission.objects.filter(
            content_type__app_label='djemtest',
            content_type__model='olptest'
        )
        
        user.user_permissions.set(permissions.filter(codename__in=('add_olptest', 'change_olptest')))
        
        group = Group.objects.create(name='Test Group')
        group.permissions.set(permissions.filter(codename='open_olptest'))
        user.groups.add(group)
        
        # OLPMixin defines the cache attribute by default, but it should be empty
        self.assertEqual(user._olp_cache, {})
        
        # Django's MLP cache attributes should not exist
        with self.assertRaises(AttributeError):
            user._user_perm_cache  # noqa: B018
        
        with self.assertRaises(AttributeError):
            user._group_perm_cache  # noqa: B018
        
        with self.assertRaises(AttributeError):
            user._perm_cache  # noqa: B018
        
        # Check a model-level permission - the MLP caches only should populate
        user.has_perm('djemtest.open_olptest')
        
        self.assertEqual(user._olp_cache, {})
        self.assertEqual(len(user._user_perm_cache), 2)
        self.assertEqual(len(user._group_perm_cache), 1)
        self.assertEqual(len(user._perm_cache), 3)
        
        # Clear the cache - Django's MLP cache attributes should be removed
        user.clear_perm_cache()
        
        with self.assertRaises(AttributeError):
            user._user_perm_cache  # noqa: B018
        
        with self.assertRaises(AttributeError):
            user._group_perm_cache  # noqa: B018
        
        with self.assertRaises(AttributeError):
            user._perm_cache  # noqa: B018
    
    def test_clear_perm_cache__olp(self):
        """
        Test the clear_perm_cache() method after object-level permissions have
        been checked. It should empty the dictionary used as the OLP cache, and
        remove the attributes used by Django for various MLP caches.
        """
        
        user = self.user
        
        # Grant the user some permissions - both directly and via a group
        permissions = Permission.objects.filter(
            content_type__app_label='djemtest',
            content_type__model='olptest'
        )
        
        user.user_permissions.set(permissions.filter(codename__in=('add_olptest', 'change_olptest')))
        
        group = Group.objects.create(name='Test Group')
        group.permissions.set(permissions.filter(codename='open_olptest'))
        user.groups.add(group)
        
        # OLPMixin defines the cache attribute by default, but it should be empty
        self.assertEqual(user._olp_cache, {})
        
        # Django's MLP cache attributes should not exist
        with self.assertRaises(AttributeError):
            user._user_perm_cache  # noqa: B018
        
        with self.assertRaises(AttributeError):
            user._group_perm_cache  # noqa: B018
        
        with self.assertRaises(AttributeError):
            user._perm_cache  # noqa: B018
        
        # Check an object-level permission - the OLP and MLP caches should populate
        obj = OLPTest.objects.create()
        user.has_perm('djemtest.open_olptest', obj)
        
        self.assertEqual(len(user._olp_cache), 1)
        self.assertEqual(len(user._user_perm_cache), 2)
        self.assertEqual(len(user._group_perm_cache), 1)
        self.assertEqual(len(user._perm_cache), 3)
        
        # Clear the cache - the OLP cache dictionary should be emptied and the
        # Django's MLP cache attributes should be removed
        user.clear_perm_cache()
        
        self.assertEqual(user._olp_cache, {})
        
        with self.assertRaises(AttributeError):
            user._user_perm_cache  # noqa: B018
        
        with self.assertRaises(AttributeError):
            user._group_perm_cache  # noqa: B018
        
        with self.assertRaises(AttributeError):
            user._perm_cache  # noqa: B018
    
    @override_settings(DJEM_PERM_LOG_VERBOSITY=0)
    def test_has_perm__logging__0(self):
        """
        Test the overridden has_perm() method when the DJEM_PERM_LOG_VERBOSITY
        setting is 0 (no logging). No log should be created automatically.
        """
        
        user = self.user
        
        user.has_perm('djemtest.mlp_log')
        
        with self.assertRaisesMessage(KeyError, 'No finished logs to retrieve'):
            user.get_last_log()
    
    @override_settings(DJEM_PERM_LOG_VERBOSITY=1)
    def test_has_perm__logging__1__mlp(self):
        """
        Test the overridden has_perm() method for a model-level check when the
        DJEM_PERM_LOG_VERBOSITY setting is 1 (automatic logging, minimal output).
        A log should be created automatically with a minimum of automatic log
        entries.
        """
        
        user = self.user
        
        user.has_perm('djemtest.mlp_log')
        
        # Only one log - that for the model-level check - should be created
        self.assertEqual(len(user._finished_logs), 1)
        
        log = user.get_last_log(raw=True)
        
        # Use repr() to verify the log lines are tagged appropriately
        self.assertEqual([repr(line) for line in log], [
            "'\\nRESULT: Permission Denied', tags=auto"
        ])
    
    @override_settings(DJEM_PERM_LOG_VERBOSITY=1)
    def test_has_perm__logging__1__olp(self):
        """
        Test the overridden has_perm() method for an object-level check when the
        DJEM_PERM_LOG_VERBOSITY setting is 1 (automatic logging, minimal output).
        A log should be created automatically with a minimum of automatic log
        entries.
        """
        
        user = self.user
        
        # Grant the user the model-level permission so that object-level checks
        # are performed
        user.user_permissions.add(Permission.objects.get(codename='olp_log'))
        
        obj = UserLogTest.objects.create()
        
        user.has_perm('djemtest.olp_log', obj)
        
        # Two logs should be created - one for the model-level permission
        # check and another for the object-level permission check
        self.assertEqual(len(user._finished_logs), 2)
        self.assertEqual(
            list(user._finished_logs.keys()),
            ['auto-djemtest.olp_log', f'auto-djemtest.olp_log-{obj.pk}']
        )
        
        log = user.get_last_log(raw=True)
        
        # Use repr() to verify the log lines are tagged appropriately
        self.assertEqual([repr(line) for line in log], [
            "'Model-level Result: Granted\\n', tags=auto",
            "'This permission is restricted.'",  # untagged
            "'\\nRESULT: Permission Denied', tags=auto"
        ])
    
    @override_settings(DJEM_PERM_LOG_VERBOSITY=1)
    def test_has_perm__logging__1__mlp__superuser(self):
        """
        Test the overridden has_perm() method for a model-level check when the
        DJEM_PERM_LOG_VERBOSITY setting is 1 (automatic logging, minimal output)
        and when the user is a superuser. A log should be created automatically
        with a minimum of automatic log entries.
        """
        
        user = self.user
        user.is_superuser = True
        user.save()
        
        user.has_perm('djemtest.mlp_log')
        
        # Only one log - that for the model-level check - should be created
        self.assertEqual(len(user._finished_logs), 1)
        
        log = user.get_last_log(raw=True)
        
        # Use repr() to verify the log lines are tagged appropriately
        self.assertEqual([repr(line) for line in log], [
            "'Active superuser: Implicit permission', tags=auto",
            "'\\nRESULT: Permission Granted', tags=auto"
        ])
    
    @override_settings(DJEM_PERM_LOG_VERBOSITY=1)
    def test_has_perm__logging__1__olp__superuser(self):
        """
        Test the overridden has_perm() method for an object-level check when the
        DJEM_PERM_LOG_VERBOSITY setting is 1 (automatic logging, minimal output)
        and when the user is a superuser. A log should be created automatically
        with a minimum of automatic log entries.
        """
        
        user = self.user
        user.is_superuser = True
        user.save()
        
        obj = UserLogTest.objects.create()
        
        user.has_perm('djemtest.olp_log', obj)
        
        # Only one log should be created - the object-level permission should
        # be implicitly granted without even needing the model-level check first
        self.assertEqual(len(user._finished_logs), 1)
        
        log = user.get_last_log(raw=True)
        
        # Use repr() to verify the log lines are tagged appropriately
        self.assertEqual([repr(line) for line in log], [
            "'Active superuser: Implicit permission', tags=auto",
            "'\\nRESULT: Permission Granted', tags=auto"
        ])
    
    @override_settings(DJEM_PERM_LOG_VERBOSITY=1, DJEM_UNIVERSAL_OLP=True)
    def test_has_perm__logging__1__mlp__superuser__universal_olp(self):
        """
        Test the overridden has_perm() method for a model-level check when the
        DJEM_PERM_LOG_VERBOSITY setting is 1 (automatic logging, minimal output),
        DJEM_UNIVERSAL_OLP is True, and when the user is a superuser. A log
        should be created automatically with a minimum of automatic log entries.
        """
        
        user = self.user
        user.is_superuser = True
        user.save()
        
        user.has_perm('djemtest.mlp_log')
        
        # Only one log - that for the model-level check - should be created
        self.assertEqual(len(user._finished_logs), 1)
        
        log = user.get_last_log(raw=True)
        
        # Use repr() to verify the log lines are tagged appropriately
        self.assertEqual([repr(line) for line in log], [
            "'Active superuser: Implicit permission (model-level)', tags=auto",
            "'\\nRESULT: Permission Granted', tags=auto"
        ])
    
    @override_settings(DJEM_PERM_LOG_VERBOSITY=1, DJEM_UNIVERSAL_OLP=True)
    def test_has_perm__logging__1__olp__superuser__universal_olp(self):
        """
        Test the overridden has_perm() method for an object-level check when the
        DJEM_PERM_LOG_VERBOSITY setting is 1 (automatic logging, minimal output),
        DJEM_UNIVERSAL_OLP is True, and when the user is a superuser. A log
        should be created automatically with a minimum of automatic log entries.
        """
        
        user = self.user
        user.is_superuser = True
        user.save()
        
        obj = UserLogTest.objects.create()
        
        user.has_perm('djemtest.olp_log', obj)
        
        # Two logs should be created - one for the model-level permission
        # check and another for the object-level permission check (which is
        # still performed for superusers with universal OLP)
        self.assertEqual(len(user._finished_logs), 2)
        self.assertEqual(
            list(user._finished_logs.keys()),
            ['auto-djemtest.olp_log', f'auto-djemtest.olp_log-{obj.pk}']
        )
        
        log = user.get_last_log(raw=True)
        
        # Use repr() to verify the log lines are tagged appropriately
        self.assertEqual([repr(line) for line in log], [
            "'Active superuser: Implicit permission (model-level)', tags=auto",
            "'Model-level Result: Granted\\n', tags=auto",
            "'This permission is restricted.'",  # untagged
            "'\\nRESULT: Permission Denied', tags=auto"
        ])
    
    @override_settings(DJEM_PERM_LOG_VERBOSITY=2)
    def test_has_perm__logging__2__mlp(self):
        """
        Test the overridden has_perm() method for a model-level check when the
        DJEM_PERM_LOG_VERBOSITY setting is 2 (automatic logging, standard output).
        A log should be created automatically with all standard automatic log
        entries.
        """
        
        user = self.user
        
        user.has_perm('djemtest.mlp_log')
        
        # Only one log - thet for the model-level check - should be created
        self.assertEqual(len(user._finished_logs), 1)
        
        log = user.get_last_log(raw=True)
        
        # Use repr() to verify the log lines are tagged appropriately
        self.assertEqual([repr(line) for line in log], [
            "'Permission: djemtest.mlp_log', tags=auto",
            f"'User: test.user ({self.user.pk})', tags=auto",
            "'', tags=auto",
            "'\\nRESULT: Permission Denied', tags=auto"
        ])
    
    @override_settings(DJEM_PERM_LOG_VERBOSITY=2)
    def test_has_perm__logging__2__olp(self):
        """
        Test the overridden has_perm() method for an object-level check when the
        DJEM_PERM_LOG_VERBOSITY setting is 2 (automatic logging, standard output).
        A log should be created automatically with all standard automatic log
        entries.
        """
        
        user = self.user
        
        # Grant the user the model-level permission so that object-level checks
        # are performed
        user.user_permissions.add(Permission.objects.get(codename='olp_log'))
        
        obj = UserLogTest.objects.create()
        
        user.has_perm('djemtest.olp_log', obj)
        
        # Two logs should be created - one for the model-level permission
        # check and another for the object-level permission check
        self.assertEqual(len(user._finished_logs), 2)
        self.assertEqual(
            list(user._finished_logs.keys()),
            ['auto-djemtest.olp_log', f'auto-djemtest.olp_log-{obj.pk}']
        )
        
        log = user.get_last_log(raw=True)
        
        # Use repr() to verify the log lines are tagged appropriately
        self.assertEqual([repr(line) for line in log], [
            "'Permission: djemtest.olp_log', tags=auto",
            f"'User: test.user ({self.user.pk})', tags=auto",
            f"'Object: Log Test #{obj.pk} ({obj.pk})', tags=auto",
            "'', tags=auto",
            "'Model-level Result: Granted\\n', tags=auto",
            "'This permission is restricted.'",  # untagged
            "'\\nRESULT: Permission Denied', tags=auto"
        ])
    
    @override_settings(DJEM_PERM_LOG_VERBOSITY=2)
    def test_has_perm__logging__2__mlp__superuser(self):
        """
        Test the overridden has_perm() method for a model-level check when the
        DJEM_PERM_LOG_VERBOSITY setting is 2 (automatic logging, standard output)
        and when the user is a superuser. A log should be created automatically
        with all standard automatic log entries.
        """
        
        user = self.user
        user.is_superuser = True
        user.save()
        
        user.has_perm('djemtest.mlp_log')
        
        # Only one log - that for the model-level check - should be created
        self.assertEqual(len(user._finished_logs), 1)
        
        log = user.get_last_log(raw=True)
        
        # Use repr() to verify the log lines are tagged appropriately
        self.assertEqual([repr(line) for line in log], [
            "'Permission: djemtest.mlp_log', tags=auto",
            f"'User: test.user ({self.user.pk})', tags=auto",
            "'', tags=auto",
            "'Active superuser: Implicit permission', tags=auto",
            "'\\nRESULT: Permission Granted', tags=auto"
        ])
    
    @override_settings(DJEM_PERM_LOG_VERBOSITY=2)
    def test_has_perm__logging__2__olp__superuser(self):
        """
        Test the overridden has_perm() method for an object-level check when the
        DJEM_PERM_LOG_VERBOSITY setting is 2 (automatic logging, standard output)
        and when the user is a superuser. A log should be created automatically
        with all standard automatic log entries.
        """
        
        user = self.user
        user.is_superuser = True
        user.save()
        
        obj = UserLogTest.objects.create()
        
        user.has_perm('djemtest.olp_log', obj)
        
        # Only one log should be created - the object-level permission should
        # be implicitly granted without even needing the model-level check first
        self.assertEqual(len(user._finished_logs), 1)
        
        log = user.get_last_log(raw=True)
        
        # Use repr() to verify the log lines are tagged appropriately
        self.assertEqual([repr(line) for line in log], [
            "'Permission: djemtest.olp_log', tags=auto",
            f"'User: test.user ({self.user.pk})', tags=auto",
            f"'Object: Log Test #{obj.pk} ({obj.pk})', tags=auto",
            "'', tags=auto",
            "'Active superuser: Implicit permission', tags=auto",
            "'\\nRESULT: Permission Granted', tags=auto"
        ])
    
    @override_settings(DJEM_PERM_LOG_VERBOSITY=2, DJEM_UNIVERSAL_OLP=True)
    def test_has_perm__logging__2__mlp__superuser__universal_olp(self):
        """
        Test the overridden has_perm() method for a model-level check when the
        DJEM_PERM_LOG_VERBOSITY setting is 2 (automatic logging, standard output),
        DJEM_UNIVERSAL_OLP is True, and when the user is a superuser. A log
        should be created automatically with all standard automatic log entries.
        """
        
        user = self.user
        user.is_superuser = True
        user.save()
        
        user.has_perm('djemtest.mlp_log')
        
        # Only one log - that for the model-level check - should be created
        self.assertEqual(len(user._finished_logs), 1)
        
        log = user.get_last_log(raw=True)
        
        # Use repr() to verify the log lines are tagged appropriately
        self.assertEqual([repr(line) for line in log], [
            "'Permission: djemtest.mlp_log', tags=auto",
            f"'User: test.user ({self.user.pk})', tags=auto",
            "'', tags=auto",
            "'Active superuser: Implicit permission (model-level)', tags=auto",
            "'\\nRESULT: Permission Granted', tags=auto"
        ])
    
    @override_settings(DJEM_PERM_LOG_VERBOSITY=2, DJEM_UNIVERSAL_OLP=True)
    def test_has_perm__logging__2__olp__superuser__universal_olp(self):
        """
        Test the overridden has_perm() method for an object-level check when the
        DJEM_PERM_LOG_VERBOSITY setting is 2 (automatic logging, standard output),
        DJEM_UNIVERSAL_OLP is True, and when the user is a superuser. A log
        should be created automatically with all standard automatic log entries.
        """
        
        user = self.user
        user.is_superuser = True
        user.save()
        
        obj = UserLogTest.objects.create()
        
        user.has_perm('djemtest.olp_log', obj)
        
        # Two logs should be created - one for the model-level permission
        # check and another for the object-level permission check (which is
        # still performed for superusers with universal OLP)
        self.assertEqual(len(user._finished_logs), 2)
        self.assertEqual(
            list(user._finished_logs.keys()),
            ['auto-djemtest.olp_log', f'auto-djemtest.olp_log-{obj.pk}']
        )
        
        log = user.get_last_log(raw=True)
        
        # Use repr() to verify the log lines are tagged appropriately
        self.assertEqual([repr(line) for line in log], [
            "'Permission: djemtest.olp_log', tags=auto",
            f"'User: test.user ({self.user.pk})', tags=auto",
            f"'Object: Log Test #{obj.pk} ({obj.pk})', tags=auto",
            "'', tags=auto",
            "'Active superuser: Implicit permission (model-level)', tags=auto",
            "'Model-level Result: Granted\\n', tags=auto",
            "'This permission is restricted.'",  # untagged
            "'\\nRESULT: Permission Denied', tags=auto"
        ])
    
    @override_settings(DJEM_PERM_LOG_VERBOSITY=0)
    def test_get_all_permissions_logging__0__olp(self):
        
        user = self.user
        obj = UserLogTest.objects.create()
        
        user.get_all_permissions(obj)
        
        self.assertEqual(len(user._active_logs), 0)
        self.assertEqual(len(user._finished_logs), 0)
    
    @override_settings(DJEM_PERM_LOG_VERBOSITY=1)
    def test_get_all_permissions_logging__1__olp(self):
        
        user = self.user
        obj = UserLogTest.objects.create()
        
        user.get_all_permissions(obj)
        
        # No logs should remain active
        self.assertEqual(len(user._active_logs), 0)
        
        # Finished logs should only exist for the model-level checks
        self.assertCountEqual(user._finished_logs.keys(), [
            'auto-djemtest.view_userlogtest', 'auto-djemtest.add_userlogtest',
            'auto-djemtest.change_userlogtest', 'auto-djemtest.delete_userlogtest',
            'auto-djemtest.mlp_log', 'auto-djemtest.olp_log'
        ])
    
    @override_settings(DJEM_PERM_LOG_VERBOSITY=2)
    def test_get_all_permissions_logging__2__olp(self):
        
        user = self.user
        obj = UserLogTest.objects.create()
        
        user.get_all_permissions(obj)
        
        # No logs should remain active
        self.assertEqual(len(user._active_logs), 0)
        
        # Finished logs should only exist for the model-level checks
        self.assertCountEqual(user._finished_logs.keys(), [
            'auto-djemtest.view_userlogtest', 'auto-djemtest.add_userlogtest',
            'auto-djemtest.change_userlogtest', 'auto-djemtest.delete_userlogtest',
            'auto-djemtest.mlp_log', 'auto-djemtest.olp_log'
        ])


@override_settings(AUTH_USER_MODEL='auth.User', AUTHENTICATION_BACKENDS=_backends)
class OLPTestCase(TestCase):
    
    UserModel = User
    TestModel = OLPTest
    model_name = 'olptest'
    
    def setUp(self):
        
        group1 = Group.objects.create(name='Test Group 1')
        group2 = Group.objects.create(name='Test Group 2')
        
        user1 = self.UserModel.objects.create_user('test1')
        user1.groups.add(group1)
        
        user2 = self.UserModel.objects.create_user('test2')
        user2.groups.add(group2)
        
        # Grant both users and both groups all permissions for OLPTest, at
        # the model level (except "closed", only accessible to super users)
        permissions = Permission.objects.filter(
            content_type__app_label='djemtest',
            content_type__model=self.model_name
        ).exclude(codename=f'closed_{self.model_name}')
        
        user1.user_permissions.set(permissions)
        user2.user_permissions.set(permissions)
        group1.permissions.set(permissions)
        group2.permissions.set(permissions)
        
        self.user1 = user1
        self.user2 = user2
        self.group1 = group1
        self.group2 = group2
        self.all_permissions = permissions
    
    def perm(self, perm_name):
        
        return f'djemtest.{perm_name}_{self.model_name}'
    
    def cache(self, cache_type, perm_name, obj):
        
        return f'{cache_type}-djemtest.{perm_name}_{self.model_name}-{obj.pk}'
    
    def cache_empty_test(self, user):
        
        # The cache attribute on the user does not exist by default
        with self.assertRaises(AttributeError):
            user._olp_cache  # noqa: B018
    
    def cache_reset_test(self, user):
        
        # Requerying for the user is the only way to reset the cache
        user = self.UserModel.objects.get(pk=user.pk)
        
        self.cache_empty_test(user)
    
    def test_auth__valid(self):
        """
        Test the backend does not interfere with valid user authentication.
        """
        
        user = self.user1
        user.set_password('blahblahblah')
        user.save()
        
        self.assertTrue(authenticate(username='test1', password='blahblahblah'))  # noqa: S106
    
    def test_auth__invalid(self):
        """
        Test the backend does not interfere with invalid user authentication.
        """
        
        self.assertFalse(authenticate(username='test1', password='badpassword'))  # noqa: S106
    
    def test_has_perm__no_model_level(self):
        """
        Test a user is denied object-level permissions if they don't have the
        corresponding model-level permissions.
        """
        
        obj = self.TestModel.objects.create()
        
        user1 = self.UserModel.objects.create_user('useless')
        perm1 = user1.has_perm(self.perm('open'), obj)
        self.assertFalse(perm1)
        
        user2 = self.UserModel.objects.create_user('useful')
        user2.user_permissions.add(Permission.objects.get(codename=f'open_{self.model_name}'))
        perm2 = user2.has_perm(self.perm('open'), obj)
        self.assertTrue(perm2)
    
    def test_has_perm__inactive_user(self):
        """
        Test an inactive user is denied object-level permissions without ever
        reaching the object's permission access method.
        """
        
        user = self.UserModel.objects.create_user('inactive')
        user.is_active = False
        user.save()
        
        # Grant the user the "open" permission to ensure it is their
        # inactive-ness that denies them permission
        user.user_permissions.add(Permission.objects.get(codename=f'open_{self.model_name}'))
        
        obj = self.TestModel.objects.create()
        
        perm = user.has_perm(self.perm('open'), obj)
        
        self.assertFalse(perm)
    
    def test_has_perm__super_user(self):
        """
        Test a superuser is granted object-level permissions without ever
        reaching the object's permission access method.
        """
        
        user = self.UserModel.objects.create_user('super')
        user.is_superuser = True
        user.save()
        
        # Deliberately do not grant the user the "open" permission to
        # ensure it is their super-ness that grants them permission
        
        obj = self.TestModel.objects.create()
        
        perm = user.has_perm(self.perm('open'), obj)
        
        self.assertTrue(perm)
    
    def test_has_perm__inactive_super_user(self):
        """
        Test an inactive superuser is denied object-level permissions without
        ever reaching the object's permission access method, despite being a
        superuser.
        """
        
        user = self.UserModel.objects.create_user('superinactive')
        user.is_superuser = True
        user.is_active = False
        user.save()
        
        # Grant the user the "open" permission to ensure it is their
        # inactive-ness that denies them permission
        user.user_permissions.add(Permission.objects.get(codename=f'open_{self.model_name}'))
        
        obj = self.TestModel.objects.create()
        
        perm = user.has_perm(self.perm('open'), obj)
        
        self.assertFalse(perm)
    
    def test_has_perm__no_access_fn(self):
        """
        Test that the lack of defined permission access methods on the object
        being tested does not deny access (checking a permission without passing
        an object should be identical to checking a permission with passing an
        object, if there is no object-level logic involved in granting/denying
        the permission).
        """
        
        obj = self.TestModel.objects.create()
        
        # Test without object
        model_perm = self.user1.has_perm(self.perm('add'))
        self.assertTrue(model_perm)
        
        # Test with object
        obj_perm = self.user1.has_perm(self.perm('add'), obj)
        self.assertTrue(obj_perm)
    
    def test_has_perm__user_only_logic(self):
        """
        Test that an object's user-based permission access method can be used
        to grant permissions to some users and not others, when all users have
        the same permission at the model level, by returning True/False, and
        that the result is unaffected by having no group-based logic defined.
        """
        
        obj = self.TestModel.objects.create(user=self.user1)
        
        perm1 = self.user1.has_perm(self.perm('user_only'), obj)
        perm2 = self.user2.has_perm(self.perm('user_only'), obj)
        
        self.assertTrue(perm1)
        self.assertFalse(perm2)
    
    def test_has_perm__group_only_logic(self):
        """
        Test that an object's group-based permission access method can be used
        to grant permissions to some users and not others based on their
        groups, when all groups have the same permission at the model level, and
        that the result is unaffected by having no user-based logic defined.
        """
        
        obj = self.TestModel.objects.create(group=self.group1)
        
        perm1 = self.user1.has_perm(self.perm('group_only'), obj)
        perm2 = self.user2.has_perm(self.perm('group_only'), obj)
        
        self.assertTrue(perm1)
        self.assertFalse(perm2)
    
    def test_has_perm__combined_logic(self):
        """
        Test that an object's user-based AND group-based permission access
        methods can be used together to grant permissions to some users and not
        others, with either one able to grant the permission if the other does
        not.
        """
        
        obj1 = self.TestModel.objects.create(user=self.user1, group=self.group2)
        obj2 = self.TestModel.objects.create()
        
        # User = True, Group untested
        user_perm1 = self.user1.has_perm(self.perm('user_only'), obj1)
        group_perm1 = self.user1.has_perm(self.perm('group_only'), obj1)
        combined_perm1 = self.user1.has_perm(self.perm('combined'), obj1)
        self.assertTrue(user_perm1)
        self.assertFalse(group_perm1)
        self.assertTrue(combined_perm1)
        
        # User = False, Group = True
        user_perm2 = self.user2.has_perm(self.perm('user_only'), obj1)
        group_perm2 = self.user2.has_perm(self.perm('group_only'), obj1)
        combined_perm2 = self.user2.has_perm(self.perm('combined'), obj1)
        self.assertFalse(user_perm2)
        self.assertTrue(group_perm2)
        self.assertTrue(combined_perm2)
        
        # User = False, Group = False
        user_perm3 = self.user1.has_perm(self.perm('user_only'), obj2)
        group_perm3 = self.user1.has_perm(self.perm('group_only'), obj2)
        combined_perm3 = self.user1.has_perm(self.perm('combined'), obj2)
        self.assertFalse(user_perm3)
        self.assertFalse(group_perm3)
        self.assertFalse(combined_perm3)
    
    def test_has_perm__permissiondenied(self):
        """
        Test that an object's permission access methods can raise the
        PermissionDenied exception and have it treated as returning False.
        """
        
        obj = self.TestModel.objects.create(user=self.user1)
        
        perm = self.user1.has_perm(self.perm('deny'), obj)
        
        self.assertFalse(perm)
    
    def test_has_perm__cache(self):
        """
        Test that determining a user's object-level permission creates a cache
        on the User instance of the result, for the permission and object tested.
        """
        
        user = self.user1
        obj = self.TestModel.objects.create()
        user_perm_cache_name = self.cache('user', 'combined', obj)
        group_perm_cache_name = self.cache('group', 'combined', obj)
        
        # Test cache does not exist
        self.cache_empty_test(user)
        
        user.has_perm(self.perm('combined'), obj)
        
        # Test cache has been set
        self.assertFalse(user._olp_cache[user_perm_cache_name])
        self.assertFalse(user._olp_cache[group_perm_cache_name])
        
        # Test resetting the cache
        self.cache_reset_test(user)
    
    def test_has_perms(self):
        """
        Test PermissionsMixin.has_perms works and correctly identifies the
        object-level permissions the user has.
        """
        
        obj = self.TestModel.objects.create(user=self.user1)
        
        perm1 = self.user1.has_perms((self.perm('open'), self.perm('combined')), obj)
        self.assertTrue(perm1)
        
        perm2 = self.user2.has_perms((self.perm('open'), self.perm('combined')), obj)
        self.assertFalse(perm2)
    
    def test_get_user_permissions(self):
        """
        Test ObjectPermissionsBackend.get_user_permissions() works and correctly
        identifies the object-level permissions the user has.
        Test the backend directly, without going through User/PermissionsMixin
        as they don't provide a mapping through to it.
        """
        
        backend = ObjectPermissionsBackend()
        obj = self.TestModel.objects.create(user=self.user1, group=self.group1)
        
        self.assertEqual(backend.get_user_permissions(self.user1, obj), {
            self.perm('view'),
            self.perm('delete'),
            self.perm('change'),
            self.perm('add'),
            self.perm('open'),
            self.perm('user_only'),
            self.perm('combined')
        })
        
        self.assertEqual(backend.get_user_permissions(self.user2, obj), {
            self.perm('view'),
            self.perm('delete'),
            self.perm('change'),
            self.perm('add'),
            self.perm('open')
        })
    
    def test_get_user_permissions__inactive_user(self):
        """
        Test ObjectPermissionsBackend.get_user_permissions() correctly denies
        all permissions to inactive users.
        Test the backend directly, without going through User/PermissionsMixin
        as they don't provide a mapping through to it.
        """
        
        backend = ObjectPermissionsBackend()
        user = self.UserModel.objects.create_user('inactive')
        user.is_active = False
        user.save()
        
        # Give the user all model-level permissions to ensure it is the
        # inactive-ness that denies them permission
        user.user_permissions.set(self.all_permissions)
        
        obj = self.TestModel.objects.create()
        
        perms = backend.get_user_permissions(user, obj)
        self.assertEqual(perms, set())
    
    def test_get_user_permissions__super_user(self):
        """
        Test ObjectPermissionsBackend.get_user_permissions() correctly grants
        all permissions to superusers.
        Test the backend directly, without going through User/PermissionsMixin
        as they don't provide a mapping through to it.
        """
        
        backend = ObjectPermissionsBackend()
        user = self.UserModel.objects.create_user('super')
        user.is_superuser = True
        user.save()
        
        # The user deliberately does not have any model-level permissions to
        # ensure it is the super-ness that grants them permission
        
        obj = self.TestModel.objects.create()
        
        perms = backend.get_user_permissions(user, obj)
        self.assertEqual(perms, {
            self.perm('view'),
            self.perm('delete'),
            self.perm('change'),
            self.perm('add'),
            self.perm('open'),
            self.perm('closed'),
            self.perm('user_only'),
            self.perm('group_only'),
            self.perm('combined'),
            self.perm('deny')
        })
    
    def test_get_user_permissions__inactive_super_user(self):
        """
        Test ObjectPermissionsBackend.get_user_permissions() correctly denies
        all permissions to inactive users, even superusers.
        Test the backend directly, without going through User/PermissionsMixin
        as they don't provide a mapping through to it.
        """
        
        backend = ObjectPermissionsBackend()
        user = self.UserModel.objects.create_user('superinactive')
        user.is_superuser = True
        user.is_active = False
        user.save()
        
        # Give the user all model-level permissions to ensure it is the
        # inactive-ness that denies them permission
        user.user_permissions.set(self.all_permissions)
        
        obj = self.TestModel.objects.create()
        
        perms = backend.get_user_permissions(user, obj)
        self.assertEqual(perms, set())
    
    def test_get_user_permissions__cache(self):
        """
        Test that PermissionsMixin.get_user_permissions correctly creates a
        a cache on the User instance of the result of each permission test, for
        the permission and object tested.
        """
        
        backend = ObjectPermissionsBackend()
        user = self.user1
        obj = self.TestModel.objects.create()
        
        expected_caches = (
            self.cache('user', 'view', obj),
            self.cache('user', 'add', obj),
            self.cache('user', 'change', obj),
            self.cache('user', 'delete', obj),
            self.cache('user', 'open', obj),
            self.cache('user', 'user_only', obj),
            self.cache('user', 'group_only', obj),
            self.cache('user', 'combined', obj),
            self.cache('user', 'deny', obj)
        )
        
        unexpected_caches = (
            self.cache('user', 'closed', obj),  # does not reach OLP stage (MLP denied)
        )
        
        unexpected_caches = [s.format(obj.pk) for s in unexpected_caches]
        
        # Test cache does not exist
        self.cache_empty_test(user)
        
        backend.get_user_permissions(user, obj)
        
        # Test expected caches have been set
        for cache_key in expected_caches:
            try:
                user._olp_cache[cache_key]
            except (AttributeError, KeyError):  # pragma: no cover # noqa: PERF203
                self.fail(f'Cache not set: {cache_key}')
        
        # Test unexpected caches still do not exist
        for cache_key in unexpected_caches:
            with self.assertRaises(KeyError):
                user._olp_cache[cache_key]
        
        # Test resetting the cache
        self.cache_reset_test(user)
    
    def test_get_group_permissions(self):
        """
        Test PermissionsMixin.get_group_permissions() works and correctly
        identifies the object-level permissions the user has.
        """
        
        obj = self.TestModel.objects.create(user=self.user1, group=self.group1)
        
        self.assertEqual(self.user1.get_group_permissions(obj), {
            self.perm('view'),
            self.perm('delete'),
            self.perm('change'),
            self.perm('add'),
            self.perm('open'),
            self.perm('group_only'),
            self.perm('combined'),
            self.perm('deny')  # no group-based access method, defaults open
        })
        
        self.assertEqual(self.user2.get_group_permissions(obj), {
            self.perm('view'),
            self.perm('delete'),
            self.perm('change'),
            self.perm('add'),
            self.perm('open'),
            self.perm('deny')  # no group-based access method, defaults open
        })
    
    def test_get_group_permissions__inactive_user(self):
        """
        Test PermissionsMixin.get_group_permissions() correctly denies all
        permissions to inactive users.
        """
        
        user = self.UserModel.objects.create_user('inactive')
        user.is_active = False
        user.save()
        
        # Give the user all model-level permissions to ensure it is the
        # inactive-ness that denies them permission
        user.user_permissions.set(self.all_permissions)
        
        obj = self.TestModel.objects.create(user=user)
        
        perms = user.get_group_permissions(obj)
        self.assertEqual(perms, set())
    
    def test_get_group_permissions__super_user(self):
        """
        Test PermissionsMixin.get_group_permissions() correctly grants all
        permissions to superusers.
        """
        
        user = self.UserModel.objects.create_user('super')
        user.is_superuser = True
        user.save()
        
        # The user deliberately does not have any model-level permissions to
        # ensure it is the super-ness that grants them permission
        
        obj = self.TestModel.objects.create()
        
        self.assertEqual(user.get_all_permissions(obj), {
            self.perm('view'),
            self.perm('delete'),
            self.perm('change'),
            self.perm('add'),
            self.perm('open'),
            self.perm('closed'),
            self.perm('user_only'),
            self.perm('group_only'),
            self.perm('combined'),
            self.perm('deny')
        })
    
    def test_get_group_permissions__inactive_super_user(self):
        """
        Test PermissionsMixin.get_group_permissions() correctly denies all
        permissions to inactive users, even superusers.
        """
        
        user = self.UserModel.objects.create_user('superinactive')
        user.is_superuser = True
        user.is_active = False
        user.save()
        
        # Give the user all model-level permissions to ensure it is the
        # inactive-ness that denies them permission
        user.user_permissions.set(self.all_permissions)
        
        obj = self.TestModel.objects.create(user=user)
        
        perms = user.get_group_permissions(obj)
        self.assertEqual(perms, set())
    
    def test_get_group_permissions__cache(self):
        """
        Test that PermissionsMixin.get_group_permissions() correctly creates a
        a cache on the User instance of the result of each permission test, for
        the permission and object tested.
        """
        
        user = self.user1
        obj = self.TestModel.objects.create()
        
        expected_caches = (
            self.cache('group', 'view', obj),
            self.cache('group', 'add', obj),
            self.cache('group', 'change', obj),
            self.cache('group', 'delete', obj),
            self.cache('group', 'open', obj),
            self.cache('group', 'user_only', obj),
            self.cache('group', 'group_only', obj),
            self.cache('group', 'combined', obj),
            self.cache('group', 'deny', obj)
        )
        
        unexpected_caches = (
            self.cache('group', 'closed', obj),  # does not reach OLP stage (MLP denied)
        )
        
        unexpected_caches = [s.format(obj.pk) for s in unexpected_caches]
        
        # Test cache does not exist
        self.cache_empty_test(user)
        
        user.get_group_permissions(obj)
        
        # Test expected caches have been set
        for cache_key in expected_caches:
            try:
                user._olp_cache[cache_key]
            except (AttributeError, KeyError):  # pragma: no cover # noqa: PERF203
                self.fail(f'Cache not set: {cache_key}')
        
        # Test unexpected caches still do not exist
        for cache_key in unexpected_caches:
            with self.assertRaises(KeyError):
                user._olp_cache[cache_key]
        
        # Test resetting the cache
        self.cache_reset_test(user)
    
    def test_get_all_permissions(self):
        """
        Test PermissionsMixin.get_all_permissions() works and correctly
        identifies the object-level permissions the user has.
        """
        
        obj = self.TestModel.objects.create(user=self.user1, group=self.group1)
        
        self.assertEqual(self.user1.get_all_permissions(obj), {
            self.perm('view'),
            self.perm('delete'),
            self.perm('change'),
            self.perm('add'),
            self.perm('open'),
            self.perm('user_only'),
            self.perm('group_only'),
            self.perm('combined')
        })
        
        self.assertEqual(self.user2.get_all_permissions(obj), {
            self.perm('view'),
            self.perm('delete'),
            self.perm('change'),
            self.perm('add'),
            self.perm('open')
        })
    
    def test_get_all_permissions__inactive_user(self):
        """
        Test PermissionsMixin.get_all_permissions() correctly denies all
        permissions to inactive users.
        """
        
        user = self.UserModel.objects.create_user('inactive')
        user.is_active = False
        user.save()
        
        # Give the user all model-level permissions to ensure it is the
        # inactive-ness that denies them permission
        user.user_permissions.set(self.all_permissions)
        
        obj = self.TestModel.objects.create(user=user)
        
        perms = user.get_all_permissions(obj)
        self.assertEqual(perms, set())
    
    def test_get_all_permissions__super_user(self):
        """
        Test PermissionsMixin.get_all_permissions() correctly grants all
        permissions to superusers.
        """
        
        user = self.UserModel.objects.create_user('super')
        user.is_superuser = True
        user.save()
        
        # The user deliberately does not have any model-level permissions to
        # ensure it is the super-ness that grants them permission
        
        obj = self.TestModel.objects.create()
        
        self.assertEqual(user.get_all_permissions(obj), {
            self.perm('view'),
            self.perm('delete'),
            self.perm('change'),
            self.perm('add'),
            self.perm('open'),
            self.perm('closed'),
            self.perm('user_only'),
            self.perm('group_only'),
            self.perm('combined'),
            self.perm('deny')
        })
    
    def test_get_all_permissions__inactive_super_user(self):
        """
        Test PermissionsMixin.get_all_permissions() correctly denies all
        permissions to inactive users, even superusers.
        """
        
        user = self.UserModel.objects.create_user('superinactive')
        user.is_superuser = True
        user.is_active = False
        user.save()
        
        # Give the user all model-level permissions to ensure it is the
        # inactive-ness that denies them permission
        user.user_permissions.set(self.all_permissions)
        
        obj = self.TestModel.objects.create(user=user)
        
        perms = user.get_all_permissions(obj)
        self.assertEqual(perms, set())
    
    def test_get_all_permissions__cache(self):
        """
        Test that PermissionsMixin.get_all_permissions() correctly creates a
        a cache on the User instance of the result of each permission test, for
        the permission and object tested.
        """
        
        user = self.user1
        obj = self.TestModel.objects.create(user=user)
        
        expected_caches = (
            self.cache('user', 'add', obj),
            self.cache('group', 'add', obj),
            
            self.cache('user', 'change', obj),
            self.cache('group', 'change', obj),
            
            self.cache('user', 'delete', obj),
            self.cache('group', 'delete', obj),
            
            self.cache('user', 'open', obj),  # group won't need checking
            
            self.cache('user', 'user_only', obj),  # group won't need checking
            
            self.cache('user', 'group_only', obj),
            self.cache('group', 'group_only', obj),
            
            self.cache('user', 'combined', obj),  # group won't need checking
            
            self.cache('user', 'deny', obj),
            self.cache('group', 'deny', obj)
        )
        
        expected_caches = [s.format(obj.pk) for s in expected_caches]
        
        unexpected_caches = (
            # Does not reach OLP stage (MLP denied)
            self.cache('user', 'closed', obj),
            self.cache('group', 'closed', obj),
            
            #  Does not reach group OLP stage (user OLP granted)
            self.cache('group', 'open', obj),
            self.cache('group', 'user_only', obj),
            self.cache('group', 'combined', obj)
        )
        
        unexpected_caches = [s.format(obj.pk) for s in unexpected_caches]
        
        # Test cache does not exist
        self.cache_empty_test(user)
        
        user.get_all_permissions(obj)
        
        # Test expected caches have been set
        for cache_key in expected_caches:
            try:
                user._olp_cache[cache_key]
            except (AttributeError, KeyError):  # pragma: no cover # noqa: PERF203
                self.fail(f'Cache not set: {cache_key}')
        
        # Test unexpected caches still do not exist
        for cache_key in unexpected_caches:
            with self.assertRaises(KeyError):
                user._olp_cache[cache_key]
        
        # Test resetting the cache
        self.cache_reset_test(user)


class OLPCacheMixin:
    
    #
    # Overrides for the cache testing helper methods defined on OLPTestCase,
    # for test cases that use a user model that incorporates OLPMixin
    #
    
    def cache_empty_test(self, user):
        
        # OLPMixin defines the cache attribute by default, but it should be empty
        self.assertEqual(user._olp_cache, {})
    
    def cache_reset_test(self, user):

        # OLPMixin provides a method to clear the cache
        user.clear_perm_cache()

        self.cache_empty_test(user)


@override_settings(AUTH_USER_MODEL='djemtest.CustomUser', DJEM_UNIVERSAL_OLP=False)
class UniversalOLPFalseTestCase(OLPCacheMixin, OLPTestCase):
    
    #
    # A repeat of the object-level permissions tests for the default user model,
    # but for one incorporating OLPMixin, and with DJEM_UNIVERSAL_OLP=False.
    # Results should be identical.
    #
    
    UserModel = CustomUser
    TestModel = UniversalOLPTest
    model_name = 'universalolptest'
    

@override_settings(AUTH_USER_MODEL='djemtest.CustomUser', DJEM_UNIVERSAL_OLP=True)
class UniversalOLPTrueTestCase(OLPCacheMixin, OLPTestCase):
    
    #
    # A repeat of the object-level permissions tests for the default user model,
    # but for one incorporating OLPMixin, and with DJEM_UNIVERSAL_OLP=True.
    # Results should be identical EXCEPT for the tests involving active
    # superusers, which should actually perform object-level permissions rather
    # than unconditionally granting such users every permission.
    #
    
    UserModel = CustomUser
    TestModel = UniversalOLPTest
    model_name = 'universalolptest'
    
    def test_get_user_permissions__super_user(self):
        """
        Test ObjectPermissionsBackend.get_user_permissions() correctly subjects
        superusers to the same object-level permission logic as a standard user
        (they simply don't require the model-level permission be granted to
        them explicitly).
        Test the backend directly, without going through User/PermissionsMixin
        as they don't provide a mapping through to it.
        """
        
        backend = ObjectPermissionsBackend()
        user = self.UserModel.objects.create_user('super')
        user.is_superuser = True
        user.save()
        
        # The user deliberately does not have any model-level permissions to
        # ensure it is the super-ness that grants them model-level permission
        
        obj = self.TestModel.objects.create()
        
        perms = backend.get_user_permissions(user, obj)
        self.assertEqual(perms, {
            self.perm('view'),
            self.perm('delete'),
            self.perm('change'),
            self.perm('add'),
            self.perm('open'),
            self.perm('closed')
        })
    
    def test_get_group_permissions__super_user(self):
        """
        Test PermissionsMixin.get_group_permissions() correctly subjects
        superusers to the same object-level permission logic as a standard user
        (they simply don't require the model-level permission be granted to
        them explicitly).
        """
        
        user = self.UserModel.objects.create_user('super')
        user.is_superuser = True
        user.save()
        
        # The user deliberately does not have any model-level permissions to
        # ensure it is the super-ness that grants them model-level permission
        
        obj = self.TestModel.objects.create()
        
        self.assertEqual(user.get_all_permissions(obj), {
            self.perm('view'),
            self.perm('delete'),
            self.perm('change'),
            self.perm('add'),
            self.perm('open'),
            self.perm('closed')
        })
    
    def test_get_all_permissions__super_user(self):
        """
        Test PermissionsMixin.get_all_permissions() correctly subjects
        superusers to the same object-level permission logic as a standard user
        (they simply don't require the model-level permission be granted to
        them explicitly).
        """
        
        user = self.UserModel.objects.create_user('super')
        user.is_superuser = True
        user.save()
        
        # The user deliberately does not have any model-level permissions to
        # ensure it is the super-ness that grants them model-level permission
        
        obj = self.TestModel.objects.create()
        
        self.assertEqual(user.get_all_permissions(obj), {
            self.perm('view'),
            self.perm('delete'),
            self.perm('change'),
            self.perm('add'),
            self.perm('open'),
            self.perm('closed')
        })


@override_settings(AUTHENTICATION_BACKENDS=_backends)
class PermissionRequiredDecoratorTestCase(TestCase):
    
    #
    # The impact of altering the DJEM_DEFAULT_403 setting cannot be tested as
    # it is read at time of import of permission_required, so any test-based
    # setting override is not recognised.
    #
    
    def setUp(self):
        
        user = get_user_model().objects.create_user('test1')
        
        # Only grant a limited subset of permissions to test when model-level
        # permissions are NOT granted
        permissions = Permission.objects.filter(
            content_type__app_label='djemtest',
            content_type__model='olptest',
            codename__in=('open_olptest', 'combined_olptest')
        )
        
        user.user_permissions.set(permissions)
        
        self.user = user
        self.olptest_with_access = OLPTest.objects.create(user=user)
        self.olptest_without_access = OLPTest.objects.create()
        self.factory = RequestFactory()
        self.resolved_login_url = resolve_url(settings.LOGIN_URL)
    
    def test_unauthenticated(self):
        """
        Test the permission_required decorator with an unauthenticated user.
        Ensure the decorator correctly redirects to the login url.
        """
        
        view = permission_required(
            'djemtest.open_olptest'
        )(_test_view)
        
        request = self.factory.get('/test/')
        request.user = AnonymousUser()
        
        response = view(request, obj=self.olptest_with_access.pk)
        
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, f'{self.resolved_login_url}?next=/test/')
    
    def test_string_arg__access(self):
        """
        Test the permission_required decorator with a valid permission as a
        single string argument.
        Ensure the decorator correctly allows access to the view for a user
        that has been granted that permission at the model level.
        """
        
        view = permission_required(
            'djemtest.open_olptest'
        )(_test_view)
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        response = view(request)
        
        self.assertContains(response, 'success', status_code=200)
    
    def test_string_arg__no_access__redirect(self):
        """
        Test the permission_required decorator with a valid permission as a
        single string argument.
        Ensure the decorator correctly denies access to the view for a user
        that has not been granted that permission at the model level, by
        redirecting to the login page.
        """
        
        view = permission_required(
            'djemtest.add_olptest'
        )(_test_view)
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        response = view(request)
        
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, f'{self.resolved_login_url}?next=/test/')
    
    def test_string_arg__no_access__redirect__custom__relative(self):
        """
        Test the permission_required decorator with a valid permission as a
        single string argument and a custom ``login_url`` is given as a
        relative url.
        Ensure the decorator correctly denies access to the view for a user
        that has not been granted that permission at the model level, by
        redirecting to a custom page specified by the decorator.
        """
        
        view = permission_required(
            'djemtest.add_olptest',
            login_url='/custom/login/'
        )(_test_view)
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        response = view(request)
        
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, '/custom/login/?next=/test/')
    
    def test_string_arg__no_access__redirect__custom__absolute(self):
        """
        Test the permission_required decorator with a valid permission as a
        single string argument and a custom ``login_url`` is given as an
        absolute url.
        Ensure the decorator correctly denies access to the view for a user
        that has not been granted that permission at the model level, by
        redirecting to a custom page specified by the decorator.
        """
        
        view = permission_required(
            'djemtest.add_olptest',
            login_url='https://example.com/custom/login/'
        )(_test_view)
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        response = view(request)
        
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response.url,
            'https://example.com/custom/login/?next=http%3A//testserver/test/'
        )
    
    def test_string_arg__no_access__raise(self):
        """
        Test the permission_required decorator with a valid permission as a
        single string argument and ``raise_exception`` given as True.
        Ensure the decorator correctly denies access to the view for a user
        that has not been granted that permission at the model level, by
        raising PermissionDenied (which would be translated into a 403 error page).
        """
        
        view = permission_required(
            'djemtest.add_olptest',
            raise_exception=True
        )(_test_view)
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        with self.assertRaises(PermissionDenied):
            view(request)
    
    def test_string_arg__invalid_perm(self):
        """
        Test the permission_required decorator with an invalid permission as a
        single string argument.
        Ensure the decorator correctly denies access to the view.
        """
        
        view = permission_required(
            'fail'
        )(_test_view)
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        response = view(request)
        
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, f'{self.resolved_login_url}?next=/test/')
    
    def test_tuple_arg__access(self):
        """
        Test the permission_required decorator with a valid permission as a
        single tuple argument.
        Ensure the decorator correctly allows access to the view for a user
        that has been granted that permission at the object level.
        """
        
        view = permission_required(
            ('djemtest.combined_olptest', 'obj')
        )(_test_view)
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        response = view(request, obj=self.olptest_with_access.pk)
        
        self.assertContains(response, 'success', status_code=200)
    
    def test_tuple_arg__no_access__redirect(self):
        """
        Test the permission_required decorator with a valid permission as a
        single tuple argument.
        Ensure the decorator correctly denies access to the view for a user
        that has not been granted that permission at the object level, by
        redirecting to the login page.
        """
        
        view = permission_required(
            ('djemtest.combined_olptest', 'obj')
        )(_test_view)
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        response = view(request, obj=self.olptest_without_access.pk)
        
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, f'{self.resolved_login_url}?next=/test/')
    
    def test_tuple_arg__no_access__redirect__custom(self):
        """
        Test the permission_required decorator with a valid permission as a
        single tuple argument and a custom ``login_url`` given.
        Ensure the decorator correctly denies access to the view for a user
        that has not been granted that permission at the object level, by
        redirecting to a custom page specified by the decorator.
        """
        
        view = permission_required(
            ('djemtest.combined_olptest', 'obj'),
            login_url='/custom/login/'
        )(_test_view)
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        response = view(request, obj=self.olptest_without_access.pk)
        
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, '/custom/login/?next=/test/')
    
    def test_tuple_arg__no_access__raise(self):
        """
        Test the permission_required decorator with a valid permission as a
        single tuple argument and ``raise_exception`` given as True.
        Ensure the decorator correctly denies access to the view for a user
        that has not been granted that permission at the object level, by
        raising PermissionDenied (which would be translated into a 403 error page).
        """
        
        view = permission_required(
            ('djemtest.combined_olptest', 'obj'),
            raise_exception=True
        )(_test_view)
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        with self.assertRaises(PermissionDenied):
            view(request, obj=self.olptest_without_access.pk)
    
    def test_tuple_arg__invalid_perm(self):
        """
        Test the permission_required decorator with an invalid permission as a
        single tuple argument.
        Ensure the decorator correctly denies access to the view.
        """
        
        view = permission_required(
            ('fail', 'obj')
        )(_test_view)
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        response = view(request, obj=self.olptest_with_access.pk)
        
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, f'{self.resolved_login_url}?next=/test/')
    
    def test_tuple_arg__invalid_object(self):
        """
        Test the permission_required decorator with a valid permission as a
        single tuple argument.
        Ensure the decorator correctly raises a Http404 exception when an
        invalid object primary key is provided (which would be translated into
        a 404 error page).
        """
        
        view = permission_required(
            ('djemtest.combined_olptest', 'obj')
        )(_test_view)
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        with self.assertRaises(Http404):
            view(request, obj=0)
    
    def test_multiple_args__access_all(self):
        """
        Test the permission_required decorator with multiple valid permissions
        as a mixture of string and tuple arguments.
        Ensure the decorator correctly allows access to the view for a user
        that has all appropriate permissions.
        """
        
        view = permission_required(
            'djemtest.open_olptest',
            ('djemtest.combined_olptest', 'obj')
        )(_test_view)
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        response = view(request, obj=self.olptest_with_access.pk)
        
        self.assertContains(response, 'success', status_code=200)
    
    def test_multiple_args__no_access__model(self):
        """
        Test the permission_required decorator with multiple valid permissions
        as a mixture of string and tuple arguments.
        Ensure the decorator correctly denies access to the view for a user
        that has is missing one of the model-level permissions, by redirecting
        to the login page.
        """
        
        view = permission_required(
            'djemtest.add_olptest',
            ('djemtest.combined_olptest', 'obj')
        )(_test_view)
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        response = view(request, obj=self.olptest_with_access.pk)
        
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, f'{self.resolved_login_url}?next=/test/')
    
    def test_multiple_args__no_access__object(self):
        """
        Test the permission_required decorator with multiple valid permissions
        as a mixture of string and tuple arguments.
        Ensure the decorator correctly denies access to the view for a user
        that has is missing one of the object-level permissions, by redirecting
        to the login page.
        """
        
        view = permission_required(
            'djemtest.open_olptest',
            ('djemtest.combined_olptest', 'obj')
        )(_test_view)
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        response = view(request, obj=self.olptest_without_access.pk)
        
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, f'{self.resolved_login_url}?next=/test/')
    
    def test_multiple_args__no_access__custom_redirect(self):
        """
        Test the permission_required decorator with multiple valid permissions
        as a mixture of string and tuple arguments, and a custom ``login_url``
        given.
        Ensure the decorator correctly denies access to the view for a user
        that has is missing one of the object-level permissions, by redirecting
        to a custom page specified by the decorator.
        """
        
        view = permission_required(
            'djemtest.open_olptest',
            ('djemtest.combined_olptest', 'obj'),
            login_url='/custom/login/'
        )(_test_view)
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        response = view(request, obj=self.olptest_without_access.pk)
        
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, '/custom/login/?next=/test/')
    
    def test_multiple_args__no_access__raise(self):
        """
        Test the permission_required decorator with multiple valid permissions
        as a mixture of string and tuple arguments.
        Ensure the decorator correctly denies access to the view for a user
        that has is missing one of the object-level permissions, by raising
        PermissionDenied (which would be translated into a 403 error page).
        """
        
        view = permission_required(
            'djemtest.open_olptest',
            ('djemtest.combined_olptest', 'obj'),
            raise_exception=True
        )(_test_view)
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        with self.assertRaises(PermissionDenied):
            view(request, obj=self.olptest_without_access.pk)
    
    def test_multiple_args__invalid_perm(self):
        """
        Test the permission_required decorator with multiple arguments, one
        of which contains an invalid permission.
        Ensure the decorator correctly denies access to the view.
        """
        
        view = permission_required(
            'djemtest.open_olptest',
            ('fail', 'obj')
        )(_test_view)
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        response = view(request, obj=self.olptest_with_access.pk)
        
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, f'{self.resolved_login_url}?next=/test/')
    
    def test_multiple_args__invalid_object(self):
        """
        Test the permission_required decorator with multiple valid permissions
        as a mixture of string and tuple arguments.
        Ensure the decorator correctly returns a 404 error page when an invalid
        object primary key is provided (which would be translated into a 404
        error page).
        """
        
        view = permission_required(
            'djemtest.open_olptest',
            ('djemtest.combined_olptest', 'obj')
        )(_test_view)
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        with self.assertRaises(Http404):
            view(request, obj=0)


@override_settings(AUTHENTICATION_BACKENDS=_backends)
class PermissionRequiredMixinTestCase(TestCase):
    
    #
    # The impact of altering the DJEM_DEFAULT_403 setting cannot be tested as
    # it is read at time of import of PermissionRequiredMixin, so any test-based
    # setting override is not recognised.
    #
    
    def setUp(self):
        
        user = get_user_model().objects.create_user('test1')
        
        # Only grant a limited subset of permissions to test when model-level
        # permissions are NOT granted
        permissions = Permission.objects.filter(
            content_type__app_label='djemtest',
            content_type__model='olptest',
            codename__in=('open_olptest', 'combined_olptest')
        )
        
        user.user_permissions.set(permissions)
        
        self.user = user
        self.olptest_with_access = OLPTest.objects.create(user=user)
        self.olptest_without_access = OLPTest.objects.create()
        self.factory = RequestFactory()
        self.resolved_login_url = resolve_url(settings.LOGIN_URL)
    
    def test_no_permissions(self):
        """
        Test the PermissionRequiredMixin with no defined permission_required.
        Ensure the mixin raises ImproperlyConfigured.
        """
        
        view = _TestView.as_view()
        
        request = self.factory.get('/test/')
        request.user = AnonymousUser()
        
        with self.assertRaises(ImproperlyConfigured):
            view(request)
    
    def test_unauthenticated__redirect(self):
        """
        Test the PermissionRequiredMixin with an unauthenticated user.
        Ensure the mixin correctly denies access to the view (the
        unauthenticated user having no permissions), by redirecting to the
        login page.
        """
        
        view = _TestView.as_view(
            permission_required='djemtest.open_olptest'
        )
        
        request = self.factory.get('/test/')
        request.user = AnonymousUser()
        
        response = view(request, obj=self.olptest_with_access.pk)
        
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, f'{self.resolved_login_url}?next=/test/')
    
    def test_unauthenticated__redirect__custom(self):
        """
        Test the PermissionRequiredMixin with an unauthenticated user.
        Ensure the mixin correctly denies access to the view (the
        unauthenticated user having no permissions), by redirecting to a custom
        page specified by ``login_url``.
        """
        
        view = _TestView.as_view(
            permission_required='djemtest.open_olptest',
            login_url='/custom/login/'
        )
        
        request = self.factory.get('/test/')
        request.user = AnonymousUser()
        
        response = view(request, obj=self.olptest_with_access.pk)
        
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, '/custom/login/?next=/test/')
    
    def test_unauthenticated__raise(self):
        """
        Test the PermissionRequiredMixin with an unauthenticated user.
        Ensure the mixin correctly denies access to the view (the
        unauthenticated user having no permissions), by raising
        PermissionDenied (which would be translated into a 403 error page).
        """
        
        view = _TestView.as_view(
            permission_required='djemtest.open_olptest',
            raise_exception=True
        )
        
        request = self.factory.get('/test/')
        request.user = AnonymousUser()
        
        with self.assertRaises(PermissionDenied):
            view(request, obj=self.olptest_with_access.pk)
    
    def test_string_arg__access(self):
        """
        Test the PermissionRequiredMixin with a valid permission as a single
        string.
        Ensure the mixin correctly allows access to the view for a user that
        has been granted that permission at the model level.
        """
        
        view = _TestView.as_view(
            permission_required='djemtest.open_olptest'
        )
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        response = view(request)
        
        self.assertContains(response, 'success', status_code=200)
    
    def test_string_arg__no_access__raise_true(self):
        """
        Test the PermissionRequiredMixin with a valid permission as a single
        string and a ``raise_exception`` set to True.
        Ensure the mixin correctly denies access to the view for a user that
        has not been granted that permission at the model level, by raising
        PermissionDenied (which would be translated into a 403 error page).
        """
        
        view = _TestView.as_view(
            permission_required='djemtest.add_olptest',
            raise_exception=True
        )
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        with self.assertRaises(PermissionDenied):
            view(request)
    
    def test_string_arg__no_access__raise_false(self):
        """
        Test the PermissionRequiredMixin with a valid permission as a single
        string and a ``raise_exception`` set to False.
        Ensure the mixin correctly denies access to the view for a user that
        has not been granted that permission at the model level, by raising
        PermissionDenied (which would be translated into a 403 error page).
        This should happen despite ``raise_exception `` being False, due to
        the user already being authenticated.
        """
        
        view = _TestView.as_view(
            permission_required='djemtest.add_olptest',
            raise_exception=False
        )
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        with self.assertRaises(PermissionDenied):
            view(request)
    
    def test_string_arg__invalid_perm(self):
        """
        Test the PermissionRequiredMixin with an invalid permission as a single
        string.
        Ensure the mixin correctly denies access to the view.
        """
        
        view = _TestView.as_view(
            permission_required='fail'
        )
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        with self.assertRaises(PermissionDenied):
            view(request)
    
    def test_tuple_arg__access(self):
        """
        Test the PermissionRequiredMixin with a valid permission as a tuple.
        Ensure the mixin correctly allows access to the view for a user that
        has been granted that permission at the object level.
        """
        
        view = _TestView.as_view(
            permission_required=[('djemtest.combined_olptest', 'obj')]
        )
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        response = view(request, obj=self.olptest_with_access.pk)
        
        self.assertContains(response, 'success', status_code=200)
    
    def test_tuple_arg__no_access__raise_true(self):
        """
        Test the PermissionRequiredMixin with a valid permission as a tuple and
        ``raise_exception`` set to True.
        Ensure the mixin correctly denies access to the view for a user that
        has not been granted that permission at the object level, by raising
        PermissionDenied (which would be translated into a 403 error page).
        """
        
        view = _TestView.as_view(
            permission_required=[('djemtest.combined_olptest', 'obj')],
            raise_exception=True
        )
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        with self.assertRaises(PermissionDenied):
            view(request, obj=self.olptest_without_access.pk)
    
    def test_tuple_arg__no_access__raise_false(self):
        """
        Test the PermissionRequiredMixin with a valid permission as a tuple and
        ``raise_exception`` set to False .
        Ensure the mixin correctly denies access to the view for a user that
        has not been granted that permission at the object level, by raising
        PermissionDenied (which would be translated into a 403 error page).
        This should happen despite ``raise_exception `` being False, due to
        the user already being authenticated.
        """
        
        view = _TestView.as_view(
            permission_required=[('djemtest.combined_olptest', 'obj')],
            raise_exception=False
        )
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        with self.assertRaises(PermissionDenied):
            view(request, obj=self.olptest_without_access.pk)
    
    def test_tuple_arg__invalid_perm(self):
        """
        Test the PermissionRequiredMixin with an invalid permission as a tuple.
        Ensure the mixin correctly denies access to the view.
        """
        
        view = _TestView.as_view(
            permission_required=[('fail', 'obj')]
        )
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        with self.assertRaises(PermissionDenied):
            view(request, obj=self.olptest_with_access.pk)
    
    def test_tuple_arg__invalid_object(self):
        """
        Test the PermissionRequiredMixin with a valid permission as a tuple.
        Ensure the mixin correctly raises a Http404 exception when an
        invalid object primary key is provided (which would be translated into
        a 404 error page).
        """
        
        view = _TestView.as_view(
            permission_required=[('djemtest.combined_olptest', 'obj')]
        )
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        with self.assertRaises(Http404):
            view(request, obj=0)
    
    def test_multiple_args__access_all(self):
        """
        Test the PermissionRequiredMixin with multiple valid permissions as a
        mixture of strings and tuples.
        Ensure the mixin correctly allows access to the view for a user that
        has all appropriate permissions.
        """
        
        view = _TestView.as_view(
            permission_required=['djemtest.open_olptest', ('djemtest.combined_olptest', 'obj')]
        )
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        response = view(request, obj=self.olptest_with_access.pk)
        
        self.assertContains(response, 'success', status_code=200)
    
    def test_multiple_args__no_access__model(self):
        """
        Test the PermissionRequiredMixin with multiple valid permissions as a
        mixture of strings and tuples.
        Ensure the mixin correctly denies access to the view for a user that
        is missing one of the model-level permissions, by raising
        PermissionDenied (which would be translated into a 403 error page).
        """
        
        view = _TestView.as_view(
            permission_required=['djemtest.add_olptest', ('djemtest.combined_olptest', 'obj')]
        )
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        with self.assertRaises(PermissionDenied):
            view(request, obj=self.olptest_without_access.pk)
    
    def test_multiple_args__no_access__object(self):
        """
        Test the PermissionRequiredMixin with multiple valid permissions as a
        mixture of string and tuple arguments.
        Ensure the mixin correctly denies access to the view for a user that
        is missing one of the object-level permissions, by raising
        PermissionDenied (which would be translated into a 403 error page).
        """
        
        view = _TestView.as_view(
            permission_required=['djemtest.open_olptest', ('djemtest.combined_olptest', 'obj')]
        )
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        with self.assertRaises(PermissionDenied):
            view(request, obj=self.olptest_without_access.pk)
    
    def test_multiple_args__invalid_perm(self):
        """
        Test the PermissionRequiredMixin with multiple permissions as a mixture
        of strings and tuples, one of which is invalid.
        Ensure the mixin correctly denies access to the view.
        """
        
        view = _TestView.as_view(
            permission_required=['djemtest.open_olptest', ('fail', 'obj')]
        )
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        with self.assertRaises(PermissionDenied):
            view(request, obj=self.olptest_with_access.pk)
    
    def test_multiple_args__invalid_object(self):
        """
        Test the PermissionRequiredMixin with multiple valid permissions as a
        mixture of strings and tuples.
        Ensure the mixin correctly raises a Http404 exception when an
        invalid object primary key is provided (which would be translated into
        a 404 error page).
        """
        
        view = _TestView.as_view(
            permission_required=['djemtest.open_olptest', ('djemtest.combined_olptest', 'obj')]
        )
        
        request = self.factory.get('/test/')
        request.user = self.user  # simulate login
        
        with self.assertRaises(Http404):
            view(request, obj=0)



from django.contrib.auth import get_user_model
from django.core.paginator import EmptyPage
from django.test import TestCase

from djem.pagination import get_page

from .models import AuditableTest


class GetPageTestCase(TestCase):
    
    @classmethod
    def setUpTestData(cls):
        
        user = get_user_model().objects.create_user('test')
        
        for _ in range(23):
            AuditableTest().save(user)
        
        cls.object_list = AuditableTest.objects.order_by('pk')
    
    def test_per_page__none(self):
        
        with self.assertRaises(TypeError):
            get_page(1, self.object_list)
    
    def test_per_page__default(self):
        
        with self.settings(DJEM_DEFAULT_PAGE_LENGTH=10):
            page = get_page(1, self.object_list)
        
        self.assertEqual(page.number, 1)
        self.assertEqual(len(page), 10)
        self.assertEqual(page.start_index(), 1)
        self.assertEqual(page.end_index(), 10)
        self.assertEqual(page.paginator.count, 23)
        self.assertEqual(page.paginator.num_pages, 3)
    
    def test_per_page__custom(self):
        
        with self.settings(DJEM_DEFAULT_PAGE_LENGTH=10):
            page = get_page(1, self.object_list, 20)
        
        self.assertEqual(page.number, 1)
        self.assertEqual(len(page), 20)
        self.assertEqual(page.start_index(), 1)
        self.assertEqual(page.end_index(), 20)
        self.assertEqual(page.paginator.count, 23)
        self.assertEqual(page.paginator.num_pages, 2)
    
    def test_page__invalid(self):
        
        page = get_page('fail', self.object_list, 10)
        
        self.assertEqual(page.number, 1)
        self.assertEqual(len(page), 10)
        self.assertEqual(page.start_index(), 1)
        self.assertEqual(page.end_index(), 10)
        self.assertEqual(page.paginator.count, 23)
        self.assertEqual(page.paginator.num_pages, 3)
    
    def test_page__negative(self):
        
        page = get_page(-1, self.object_list, 10)
        
        self.assertEqual(page.number, 1)
        self.assertEqual(len(page), 10)
        self.assertEqual(page.start_index(), 1)
        self.assertEqual(page.end_index(), 10)
        self.assertEqual(page.paginator.count, 23)
        self.assertEqual(page.paginator.num_pages, 3)
    
    def test_page__large(self):
        
        page = get_page(10000, self.object_list, 10)
        
        self.assertEqual(page.number, 3)
        self.assertEqual(len(page), 3)
        self.assertEqual(page.start_index(), 21)
        self.assertEqual(page.end_index(), 23)
        self.assertEqual(page.paginator.count, 23)
        self.assertEqual(page.paginator.num_pages, 3)
    
    def test_empty_list__allow_empty_first__page0(self):
        
        page = get_page(0, [], 10)
        
        self.assertEqual(page.number, 1)
        self.assertEqual(len(page), 0)
        self.assertEqual(page.paginator.count, 0)
        self.assertEqual(page.paginator.num_pages, 1)
    
    def test_empty_list__allow_empty_first__page1(self):
        
        page = get_page(1, [], 10)
        
        self.assertEqual(page.number, 1)
        self.assertEqual(len(page), 0)
        self.assertEqual(page.paginator.count, 0)
        self.assertEqual(page.paginator.num_pages, 1)
    
    def test_empty_list__allow_empty_first__page2(self):
        
        page = get_page(2, [], 10)
        
        self.assertEqual(page.number, 1)
        self.assertEqual(len(page), 0)
        self.assertEqual(page.paginator.count, 0)
        self.assertEqual(page.paginator.num_pages, 1)
    
    def test_empty_list__deny_empty_first__page0(self):
        
        # Specifically ensure it is not the "That page number is less than 1" error
        with self.assertRaisesMessage(EmptyPage, 'That page contains no results'):
            get_page(0, [], 10, allow_empty_first_page=False)
    
    def test_empty_list__deny_empty_first__page1(self):
        
        # Specifically ensure it is not the "That page number is less than 1" error
        with self.assertRaisesMessage(EmptyPage, 'That page contains no results'):
            get_page(1, [], 10, allow_empty_first_page=False)
    
    def test_empty_list__deny_empty_first__page2(self):
        
        # Specifically ensure it is not the "That page number is less than 1" error
        with self.assertRaisesMessage(EmptyPage, 'That page contains no results'):
            get_page(2, [], 10, allow_empty_first_page=False)



from djem.utils.tests import setup_test_app

# Use an explicit app label so as not to interfere with the tests of
# setup_test_app() itself (which test implicit app labels)
setup_test_app(__package__, 'djemtest')  # noqa: RUF067

# Target for ROOT_URLCONF, since it has to point somewhere
urlpatterns = []  # noqa: RUF067



from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission


class Developer:
    """
    Designed for use within the Django shell to aid developers in performing
    regular operations used for testing and debugging, primarily regarding
    accessing the developer's user record and altering aspects of that record
    """
    
    awesome = {
        'is_staff': True,
        'is_superuser': True
    }
    
    boring = {
        'is_staff': False,
        'is_superuser': False
    }
    
    def __init__(self, **user_lookup_kwargs):
        
        self.user_lookup_kwargs = user_lookup_kwargs
    
    @property
    def user(self):
        """
        The developer's associated user model instance, as given by
        ``django.contrib.auth.get_user_model()``, and looked up using the
        keyword arguments given to the constructor or defined by the
        :setting:`DJEM_DEV_USER` setting.
        """
        
        if not hasattr(self, '_user'):
            if not self.user_lookup_kwargs:
                self.user_lookup_kwargs = getattr(settings, 'DJEM_DEV_USER', {})
            
            self._user = get_user_model().objects.get(**self.user_lookup_kwargs)
        
        return self._user
    
    def be_awesome(self):
        """
        Assign :attr:`user` the attributes defined in :attr:`awesome`. Designed
        to allow developers to adopt/restore relevant elevated privileges.
        """
        
        user = self.user
        
        for attr, value in self.awesome.items():
            setattr(user, attr, value)
        
        user.save(update_fields=self.awesome.keys())
    
    def be_boring(self):
        """
        Assign :attr:`user` the attributes defined in :attr:`boring`. Designed
        to allow developers to adopt relevant reduced privileges for testing.
        """
        
        user = self.user
        
        for attr, value in self.boring.items():
            setattr(user, attr, value)
        
        user.save(update_fields=self.boring.keys())
    
    def no_super(self):
        """
        Set the ``is_superuser=False`` flag on :attr:`user`.
        """
        
        user = self.user
        user.is_superuser = False
        user.save(update_fields=('is_superuser', ))
    
    def no_staff(self):
        """
        Set the ``is_staff=False`` flag on :attr:`user`.
        """
        
        user = self.user
        user.is_staff = False
        user.save(update_fields=('is_staff', ))
    
    def add_permissions(self, *permissions):
        """
        Assign :attr:`user` the given permissions (by codename).
        """
        
        permissions = Permission.objects.filter(codename__in=permissions)
        
        for perm in permissions:
            self.user.user_permissions.add(perm)
    
    def remove_permissions(self, *permissions):
        """
        Remove the given permissions (by codename) from :attr:`user`.
        """
        
        permissions = Permission.objects.filter(codename__in=permissions)
        
        for perm in permissions:
            self.user.user_permissions.remove(perm)



import inspect as inspector
import pprint
import types

from django.core.exceptions import FieldDoesNotExist
from django.db.models import Manager, Model, QuerySet
from django.db.models.fields import NOT_PROVIDED
from django.utils.encoding import force_str

from djem.utils.table import Table


def get_defined_by(obj, attr):
    """
    Return the class that defines the given attribute on the given object.
    It may be the object's own class, or one of its parents.
    """
    
    if inspector.isclass(obj):
        cls = obj
    else:
        cls = obj.__class__
    
    for parent in reversed(cls.mro()):
        if hasattr(parent, attr):
            return parent
    
    return cls


def inspectf(func):
    """
    Inspect a function, returning useful info such as its signature and where it
    is defined.
    """
    
    arguments = []
    
    argspec = inspector.getargspec(func)
    args = argspec.args
    defaults = argspec.defaults
    
    # There may be more arguments than defaults. Define an offset that will go
    # negative in that case. Increment it during the loop over args and begin
    # retrieving default values once it reaches 0.
    if not defaults:  # account for defaults being None
        offset = 0 - len(args)
    else:
        offset = len(defaults) - len(args)
    
    for arg_string in args:
        if offset >= 0:
            arg_string = f'{arg_string}={defaults[offset]}'  # noqa: PLW2901
        
        arguments.append(arg_string)
        offset += 1
    
    signature = f'{func.__name__}({", ".join(arguments)})'
    
    defined = f'Defined on line {func.__code__.co_firstlineno} of {func.__module__} ({func.__code__.co_filename})'
    
    return '\n\n'.join((signature, func.__doc__ or '[no doc string]', defined))


class InspectTable:
    
    def get_preamble(self):
        
        return None
    
    def get_data_row_count(self):
        
        return 0
    
    def get_empty_string(self):
        
        return 'No results.'
    
    def get_title(self):
        
        return None
    
    def get_description(self):
        
        return None
    
    def get_headings(self):
        
        return None
    
    def get_footer(self):
        
        return None
    
    def populate_data(self, table):
        
        raise NotImplementedError
    
    def build(self):
        
        output = [
            ''  # start with a blank line
        ]
        
        preamble = self.get_preamble()
        if preamble:
            output.extend((preamble, ''))
        
        if not self.get_data_row_count():
            output.append(self.get_empty_string())
        else:
            t = Table(
                title=self.get_title(),
                footer=self.get_footer()
            )
            
            description = self.get_description()
            if description:
                t.add_full_width_row(description, 'centre')
                t.add_row(Table.HR)
            
            # Add headings manually to put them beneath the description row/s
            headings = self.get_headings()
            if headings:
                t.add_row(headings)
            
            self.populate_data(t)
            
            output.append(t.build_table())
        
        output.append('')  # end with a blank line
        
        return '\n'.join(output)


class ObjectTable(InspectTable):
    
    def __init__(
            self,
            obj,
            ignore_methods=False,
            ignore_inherited=False,
            ignore_private=True,
            ignore_magic=True):
        
        super().__init__()
        
        self.obj = obj
        self.ignore_methods = ignore_methods
        self.ignore_inherited = ignore_inherited
        self.ignore_private = ignore_private
        self.ignore_magic = ignore_magic
        
        magic, methods, other = self._inspect_obj()
        num_inspected = 0
        data_groups = []
        
        if magic:
            num_inspected += len(magic)
            data_groups.append(magic)
        
        if methods:
            num_inspected += len(methods)
            data_groups.append(methods)
        
        if other:
            num_inspected += len(other)
            data_groups.append(other)
        
        self.num_inspected = num_inspected
        self.data_groups = data_groups
    
    def _get_inspect_value(self, v, obj, attr):
        """
        Return the given value in a format suitable for output in the table.
        """
        
        if inspector.isclass(obj):
            cls = obj
        else:
            cls = obj.__class__
        
        if isinstance(v, Manager):
            try:
                model_field = cls._meta.get_field(attr)
            except FieldDoesNotExist:
                v = f'Manager on {cls.__name__} model'
            else:
                v = f'Referencing {v.count()} {model_field.related_model.__name__} records'
        elif isinstance(v, types.MethodType):
            v = f'pp({cls.__name__}.{attr})'
        else:
            v = force_str(v)
        
        return v
    
    def _inspect_obj(self):
        
        obj = self.obj
        cls = obj.__class__
        
        ignore_methods = self.ignore_methods
        ignore_inherited = self.ignore_inherited
        ignore_private = self.ignore_private
        ignore_magic = self.ignore_magic
        
        magic = []
        methods = []
        other = []
        
        for attr in sorted(dir(obj)):
            is_magic = attr.startswith('__') and attr.endswith('__')
            
            if is_magic and ignore_magic:
                continue
            
            if attr.startswith('_') and not is_magic and ignore_private:
                continue
            
            defined_by = get_defined_by(obj, attr).__name__
            if defined_by != cls.__name__ and ignore_inherited:
                continue
            
            try:
                v = getattr(obj, attr)
            except Exception as e:
                v = f'Error accessing attribute: {e}'
                t = 'unknown'
            else:
                t = type(v).__name__
            
            if isinstance(v, types.MethodType):
                if ignore_methods:
                    continue
                
                title = f'{attr}()'
                append_to = methods
            else:
                title = force_str(attr)
                append_to = other
            
            if is_magic:
                append_to = magic
            
            append_to.append((
                defined_by,
                title,
                t,
                self._get_inspect_value(v, obj, attr)
            ))
        
        return magic, methods, other
    
    def get_data_row_count(self):
        
        return self.num_inspected
    
    def get_title(self):
        
        return f'Inspecting {self.obj.__class__.__name__} instance'
    
    def get_description(self):
        
        ignoring = []
        if self.ignore_methods:
            ignoring.append('Methods')
        if self.ignore_magic:
            ignoring.append('Magic Stuff')
        if self.ignore_private:
            ignoring.append('Private Stuff')
        if self.ignore_inherited:
            ignoring.append('Inherited Stuff')
        
        description = []
        if ignoring:
            description.append(f'Ignoring: {", ".join(ignoring)}')
        else:
            description.append('Ignoring: Nothing')
        
        try:
            mro = self.obj.mro()
        except AttributeError:
            # An instance - get the mro of the class
            mro = self.obj.__class__.mro()
        
        description.append(f'MRO: {", ".join([c.__name__ for c in mro])}')
        
        return '\n'.join(description)
    
    def get_footer(self):
        
        num_inspected = self.num_inspected
        plural = 's' if num_inspected != 1 else ''
        
        return f'{num_inspected} attribute{plural} inspected'
    
    def populate_data(self, table):
        
        if not self.num_inspected:
            table.add_full_row('No attributes to inspect.')
            return
        
        for i, group in enumerate(self.data_groups):
            if i:
                table.add_row(Table.HR)
            
            table.add_rows(group)


class ModelTable(InspectTable):
    
    FIELD_TYPE_REPLACEMENT_MAP = {
        'Field': '',
        'ForeignKey': 'FK',
        'ManyToMany': 'M2M',
        'OneToOne': 'O2O',
        'ManyToOne': 'M2O'
    }
    
    def __init__(self, model, *field_filters, **kwargs):
        
        super().__init__()
        
        self.model = model
        self.field_filters = [f.lower() for f in field_filters]
        self.concrete_only = kwargs.pop('concrete_only', True)
        
        hierarchy, pk_hierarchy = self._get_hierarchies()
        
        self.hierarchy = hierarchy
        self.pk_hierarchy = pk_hierarchy
        
        self._discover_fields()
        
        self.total_discovered_fields = sum(c['num_discovered_fields'] for c in hierarchy)
        self.total_discovered_fields += 1  # add the pk
        
        self.total_matching_fields = sum(len(c['matching_fields']) for c in hierarchy)
    
    def _get_hierarchies(self):
        
        hierarchy = []
        pk_hierarchy = []
        
        for cls in reversed(self.model.mro()):
            try:
                meta = cls._meta
            except AttributeError:  # noqa: PERF203
                # This parent class is not a Model
                pass
            else:
                model_name = cls.__name__
                
                if meta.abstract:
                    model_name += ' [Abstract]'
                else:
                    pk_hierarchy.append(meta.pk.name)
                
                hierarchy.append({
                    'display_name': model_name,
                    'class': cls,
                    'meta': meta
                })
        
        return hierarchy, pk_hierarchy
    
    def _check_field_match(self, field):
        
        field_name = field.name.lower()
        
        return any(f in field_name for f in self.field_filters)
    
    def _get_field_flags(self, field):
        
        if not field.concrete:
            # Abstract fields do not have most of the attributes checked below,
            # nor do they apply, so return early.
            return ['Not Concrete']
        
        flags = []
        
        if field.unique:
            flags.append('Unique')
        
        if field.null and not field.blank:
            flags.append('Not blank')
        
        if not field.editable:
            flags.append('Not editable')
        
        return flags
    
    def _get_field_type(self, field):
        
        field_type = field.__class__.__name__
        
        for string, replacement in self.FIELD_TYPE_REPLACEMENT_MAP.items():
            field_type = field_type.replace(string, replacement)
        
        # Add the foreign model to the type display of relationship fields
        rel = field.related_model
        if rel:
            # If a string reference, grab the model portion (if given in
            # <app_label>.<model> format). If a model class, use its name.
            if isinstance(rel, str):
                rel = rel.split('.')[-1]
            else:
                rel = rel.__name__
            
            field_type = f'{field_type} ({rel})'
        
        if not field.concrete:
            # Stop here for non-concrete fields
            return field_type
        
        # Add the size to the type display of CharField, DecimalField, etc
        try:
            size = field.max_digits
        except AttributeError:
            size = field.max_length
        else:
            try:
                size = f'{size}/{field.decimal_places}'
            except AttributeError:
                pass
        
        if size:
            field_type = f'{field_type} ({size})'
        
        return field_type
    
    def _get_field_default(self, field):
        
        default_value = field.default
        
        if default_value is NOT_PROVIDED:
            default_value = ''
        elif callable(default_value):
            default_value = f'{default_value.__name__}()'
        
        return default_value
    
    def _get_field_sort_key(self, field):
        
        prefix = 0
        
        if not field.concrete:
            # E.g. Automatic reverse ForeignKey field
            prefix = 3
        elif field.many_to_many:
            prefix = 2
        elif field.is_relation:
            prefix = 1
        
        return f'{prefix}{field.name}'
    
    def _discover_fields(self):
        
        # Track fields considered valid
        valid_fields = set()
        
        for config in self.hierarchy:
            num_discovered_fields = 0
            matching_fields = []
            
            all_fields = config['meta'].get_fields(include_parents=False)
            
            for field in sorted(all_fields, key=self._get_field_sort_key):
                if getattr(field, 'primary_key', False):
                    # Exclude primary key fields. Some automatic fields do not
                    # have the primary_key attribute to check.
                    continue
                
                if self.concrete_only and not field.concrete:
                    # Exclude non-concrete fields unless they are explicitly
                    # specified to be included
                    continue
                
                if field.name in valid_fields:
                    # This field has been processed before. Though the field
                    # list will exclude fields from parent models, it only
                    # excludes those from *concrete* parents. Fields from
                    # abstract parents should also be excluded.
                    continue
                
                # Consider "valid" before filtering out
                valid_fields.add(field.name)
                
                # Count against total before filtering out
                num_discovered_fields += 1
                
                # Apply field filters, if any
                if self.field_filters and not self._check_field_match(field):
                    continue
                
                field_type = self._get_field_type(field)
                default_value = self._get_field_default(field)
                flags = self._get_field_flags(field)
                
                matching_fields.append((
                    field.name,
                    field_type,
                    field.null,
                    default_value,
                    ', '.join(flags)
                ))
            
            config['matching_fields'] = matching_fields
            config['num_discovered_fields'] = num_discovered_fields
    
    def _get_model_row(self, config):
        
        attrs = [
            config['class'].__name__
        ]
        
        if config['meta'].abstract:
            attrs.append('[Abstract]')
        
        num_discovered_fields = config['num_discovered_fields']
        num_matching_fields = len(config['matching_fields'])
        
        if num_matching_fields == num_discovered_fields:
            attrs.append(f'\tDiscovered Fields: {num_discovered_fields}')
        else:
            attrs.append(f'\tMatching Fields: {num_matching_fields}/{num_discovered_fields}')
        
        return ' '.join(attrs)
    
    def get_data_row_count(self):
        
        return self.total_matching_fields
    
    def get_empty_string(self):
        
        filters = self.field_filters
        if filters:
            return f'Given filters matched no results: {filters}'
        
        return 'No fields to display'
    
    def get_preamble(self):
        
        m = self.model
        unique_together = m._meta.unique_together or 'None'
        ordering = m._meta.ordering or 'None'
        
        hierarchy = self.hierarchy
        hierarchy_str = '\n - '.join([c['display_name'] for c in hierarchy[:-1]])
        
        pk_hierarchy = self.pk_hierarchy
        pk_hierarchy_str = ''
        if len(pk_hierarchy) > 1:
            pk_hierarchy_str = f' ({" -> ".join(pk_hierarchy)})'
        
        lines = [
            f'Model: {m.__name__}',
            f'\nInherits From:\n - {hierarchy_str}',
            f'\nUnique Together: {unique_together}',
            f'Default Ordering: {ordering}',
            f'\nPK: {m._meta.pk.name}{pk_hierarchy_str}'
        ]
        
        return '\n'.join(lines)
    
    def get_title(self):
        
        title = f'Fields on {self.model.__name__}'
        
        filters = self.field_filters
        if filters:
            filter_str = ' || '.join([repr(s) for s in filters])
            title = f'{title} (matching {filter_str})'
        
        if self.concrete_only:
            title = f'{title} [Concrete Only]'
        
        return title
    
    def get_headings(self):
        
        return ('', '', 'Null?', 'Default', 'Flags')
    
    def get_footer(self):
        
        if self.field_filters:
            return f'Fields: {self.total_matching_fields} / {self.total_discovered_fields}'
        
        return f'Fields: {self.total_discovered_fields}'
    
    def populate_data(self, table):
        
        for config in self.hierarchy:
            matching_fields = config['matching_fields']
            if matching_fields:
                model_row = self._get_model_row(config)
                
                table.add_row(Table.HR)
                table.add_row(Table.BR)
                table.add_full_width_row(model_row)
                table.add_row(Table.BR)
                table.add_row(Table.HR)
                
                table.add_rows(matching_fields)


def pp(obj, *args, **kwargs):
    """
    Catch-all pretty-print/inspection function. Takes any object and tries to
    print the most useful representation of it. Additional args/kwargs depend
    on the type of object. Accepted types:
      - dictionaries, lists, tuples, sets and Django QuerySets: print using
        Python's pprint library. Accepts all args/kwargs of pprint.pprint().
      - functions/methods: print using inspectf. Accepts all args/kwargs of inspectf.
      - Django Model classes: print in a ModelTable. Accepts all constructor
        args/kwargs of ModelTable.
      - anything else: print in an ObjectTable. Accepts all constructor
        args/kwargs of ObjectTable.
    """
    
    # Force QuerySets to lists for pretty-printing
    if isinstance(obj, QuerySet):
        obj = list(obj)
    
    if isinstance(obj, (dict, list, tuple, set)):
        pprint.pprint(obj, *args, **kwargs)
        return
    
    function_types = (
        types.FunctionType,
        types.BuiltinFunctionType,
        types.MethodType,
        types.BuiltinMethodType
    )
    
    if isinstance(obj, function_types):
        print('\n', inspectf(obj, *args, **kwargs), '\n')
        return
    
    try:
        is_model_class = issubclass(obj, Model)
    except TypeError:
        # obj not a class
        pass
    else:
        if is_model_class:
            print(ModelTable(obj, *args, **kwargs).build())
            return
    
    print(ObjectTable(obj, *args, **kwargs).build())



import datetime
import warnings
from copy import deepcopy
from unittest.mock import MagicMock, patch
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

from django.apps import apps
from django.test import SimpleTestCase, override_settings
from django.utils import timezone

from djem import UNDEFINED
from djem.utils.dev import Developer
from djem.utils.dt import TimeZoneHelper
from djem.utils.tests import setup_test_app


class UndefinedTestCase(SimpleTestCase):
    
    def test_falsey(self):
        
        self.assertFalse(UNDEFINED)
    
    def test_print(self):
        
        self.assertEqual(str(UNDEFINED), '<undefined>')
    
    def test_deepcopy__direct(self):
        
        clone = deepcopy(UNDEFINED)
        
        self.assertIs(clone, UNDEFINED)
    
    def test_deepcopy__indirect(self):
        
        val = {'value': UNDEFINED}
        clone = deepcopy(val)
        
        self.assertIs(clone['value'], UNDEFINED)


class DeveloperTestCase(SimpleTestCase):
    """
    Tests for the Developer utility class.
    """
    
    @override_settings(DJEM_DEV_USER={'pk': 1})
    def test_user__explicit_kwargs(self):
        """
        Test the `user` property when explicit lookup kwargs are provided.
        """
        
        dev = Developer(username='testuser')
        
        mock_user = MagicMock()
        with patch('djem.utils.dev.get_user_model') as mock_get_user_model:
            mock_get_user_model.return_value.objects.get.return_value = mock_user
            
            user = dev.user
            
            mock_get_user_model.return_value.objects.get.assert_called_once_with(username='testuser')
        
        self.assertIs(user, mock_user)
    
    @override_settings(DJEM_DEV_USER={'pk': 1})
    def test_user__settings(self):
        """
        Test the `user` property when no explicit lookup kwargs are provided,
        but the `DJEM_DEV_USER` setting is defined.
        """
        
        dev = Developer()
        
        mock_user = MagicMock()
        with patch('djem.utils.dev.get_user_model') as mock_get_user_model:
            mock_get_user_model.return_value.objects.get.return_value = mock_user
            
            user = dev.user
            
            mock_get_user_model.return_value.objects.get.assert_called_once_with(pk=1)
        
        self.assertIs(user, mock_user)
    
    def test_user__no_kwargs(self):
        """
        Test the `user` property when no lookup kwargs are provided, either
        explicitly or via the `DJEM_DEV_USER` setting.
        """
        
        dev = Developer()
        
        mock_user = MagicMock()
        with patch('djem.utils.dev.get_user_model') as mock_get_user_model:
            mock_get_user_model.return_value.objects.get.return_value = mock_user
            
            user = dev.user
            
            # Should be called without arguments - would error in actual use
            mock_get_user_model.return_value.objects.get.assert_called_once_with()
        
        self.assertIs(user, mock_user)
    
    def test_user__cached(self):
        """
        Test that the `user` property is cached after first access.
        """
        
        dev = Developer(username='testuser')
        
        mock_user = MagicMock()
        with patch('djem.utils.dev.get_user_model') as mock_get_user_model:
            mock_get_user_model.return_value.objects.get.return_value = mock_user
            
            user1 = dev.user
            user2 = dev.user
            
            # Should only be called once due to caching
            mock_get_user_model.return_value.objects.get.assert_called_once()
        
        self.assertIs(user1, user2)
    
    def test_be_awesome(self):
        """
        Test the be_awesome() method. It should set the fields defined by the
        `awesome` class attribute.
        """
        
        dev = Developer(username='testuser')
        
        mock_user = MagicMock()
        with patch('djem.utils.dev.get_user_model') as mock_get_user_model:
            mock_get_user_model.return_value.objects.get.return_value = mock_user
            
            dev.be_awesome()
        
        self.assertTrue(mock_user.is_staff)
        self.assertTrue(mock_user.is_superuser)
        mock_user.save.assert_called_once()
    
    def test_be_awesome__customised(self):
        """
        Test the be_awesome() method on a subclass that customises the `awesome`
        class attribute. It should set the fields defined by the customisations.
        """
        
        class CustomDeveloper(Developer):
            
            awesome = {
                'is_superuser': True,
                'is_developer': True,
                'custom_field': 'awesome_value'
            }
        
        dev = CustomDeveloper(username='testuser')
        
        mock_user = MagicMock()
        with patch('djem.utils.dev.get_user_model') as mock_get_user_model:
            mock_get_user_model.return_value.objects.get.return_value = mock_user
            
            dev.be_awesome()
        
        self.assertTrue(mock_user.is_superuser)
        self.assertTrue(mock_user.is_developer)
        self.assertEqual(mock_user.custom_field, 'awesome_value')
        mock_user.save.assert_called_once()
    
    def test_be_boring(self):
        """
        Test the be_boring() method. It should set the fields defined by the
        `boring` class attribute.
        """
        
        dev = Developer(username='testuser')
        
        mock_user = MagicMock()
        with patch('djem.utils.dev.get_user_model') as mock_get_user_model:
            mock_get_user_model.return_value.objects.get.return_value = mock_user
            
            dev.be_boring()
        
        self.assertFalse(mock_user.is_staff)
        self.assertFalse(mock_user.is_superuser)
        mock_user.save.assert_called_once()
    
    def test_be_boring__customised(self):
        """
        Test the be_boring() method on a subclass that customises the `boring`
        class attribute. It should set the fields defined by the customisations.
        """
        
        class CustomDeveloper(Developer):
            
            boring = {
                'is_superuser': False,
                'is_developer': False,
                'custom_field': 'boring_value'
            }
        
        dev = CustomDeveloper(username='testuser')
        
        mock_user = MagicMock()
        with patch('djem.utils.dev.get_user_model') as mock_get_user_model:
            mock_get_user_model.return_value.objects.get.return_value = mock_user
            
            dev.be_boring()
        
        self.assertFalse(mock_user.is_superuser)
        self.assertFalse(mock_user.is_developer)
        self.assertEqual(mock_user.custom_field, 'boring_value')
        mock_user.save.assert_called_once()
    
    def test_no_super(self):
        """
        Test the no_super() method. It should set `is_superuser` to False.
        """
        
        dev = Developer(username='testuser')
        
        mock_user = MagicMock()
        with patch('djem.utils.dev.get_user_model') as mock_get_user_model:
            mock_get_user_model.return_value.objects.get.return_value = mock_user
            
            dev.no_super()
        
        # Only `is_superuser` should be changed
        self.assertFalse(mock_user.is_superuser)
        self.assertTrue(mock_user.is_staff)
        mock_user.save.assert_called_once_with(update_fields=('is_superuser',))
    
    def test_no_staff(self):
        """
        Test the no_staff() method. It should set `is_staff` to False.
        """
        
        dev = Developer(username='testuser')
        
        mock_user = MagicMock()
        with patch('djem.utils.dev.get_user_model') as mock_get_user_model:
            mock_get_user_model.return_value.objects.get.return_value = mock_user
            
            dev.no_staff()
        
        # Only `is_staff` should be changed
        self.assertFalse(mock_user.is_staff)
        self.assertTrue(mock_user.is_superuser)
        mock_user.save.assert_called_once_with(update_fields=('is_staff',))
    
    def test_add_permissions(self):
        """
        Test the add_permissions() method. It should add the specified
        permissions to the user.
        """
        
        dev = Developer(username='testuser')
        
        mock_user = MagicMock()
        perm1 = MagicMock()
        perm2 = MagicMock()
        perm3 = MagicMock()
        
        with patch('djem.utils.dev.get_user_model') as mock_get_user_model:
            mock_get_user_model.return_value.objects.get.return_value = mock_user
            
            with patch('djem.utils.dev.Permission') as mock_permission:
                mock_permission.objects.filter.return_value = [perm1, perm2, perm3]
                
                dev.add_permissions('add_mymodel', 'change_mymodel', 'delete_mymodel')
                
                mock_permission.objects.filter.assert_called_once_with(
                    codename__in=('add_mymodel', 'change_mymodel', 'delete_mymodel')
                )
            
            mock_user.user_permissions.add.assert_any_call(perm1)
            mock_user.user_permissions.add.assert_any_call(perm2)
            mock_user.user_permissions.add.assert_any_call(perm3)
    
    def test_remove_permissions(self):
        """
        Test the remove_permissions() method. It should remove the specified
        permissions from the user.
        """
        
        dev = Developer(username='testuser')
        
        mock_user = MagicMock()
        perm1 = MagicMock()
        perm2 = MagicMock()
        perm3 = MagicMock()
        
        with patch('djem.utils.dev.get_user_model') as mock_get_user_model:
            mock_get_user_model.return_value.objects.get.return_value = mock_user
            
            with patch('djem.utils.dev.Permission') as mock_permission:
                mock_permission.objects.filter.return_value = [perm1, perm2, perm3]
                
                dev.remove_permissions('add_mymodel', 'change_mymodel', 'delete_mymodel')
                
                mock_permission.objects.filter.assert_called_once_with(
                    codename__in=('add_mymodel', 'change_mymodel', 'delete_mymodel')
                )
            
            mock_user.user_permissions.remove.assert_any_call(perm1)
            mock_user.user_permissions.remove.assert_any_call(perm2)
            mock_user.user_permissions.remove.assert_any_call(perm3)


class TimeZoneHelperTestCase(SimpleTestCase):
    
    def test_init__good_string(self):
        """
        Test constructing a TimeZoneHelper object with a valid timezone string.
        """
        
        helper = TimeZoneHelper('Australia/Sydney')
        
        self.assertEqual(helper.tz.key, 'Australia/Sydney')
    
    def test_init__bad_string(self):
        """
        Test constructing a TimeZoneHelper object with an invalid timezone string.
        """
        
        with self.assertRaises(ZoneInfoNotFoundError):
            TimeZoneHelper('fail')
    
    def test_init__utc(self):
        """
        Test constructing a TimeZoneHelper object with the UTC singleton.
        """
        
        tz = datetime.timezone.utc
        helper = TimeZoneHelper(tz)
        
        self.assertIs(helper.tz, tz)
    
    def test_init__timezone(self):
        """
        Test constructing a TimeZoneHelper object with a valid ZoneInfo
        instance.
        """
        
        tz = ZoneInfo('Australia/Sydney')
        helper = TimeZoneHelper(tz)
        
        self.assertIs(helper.tz, tz)
    
    def test_name(self):
        """
        Test the TimeZoneHelper ``name`` property for several different
        timezones.
        """
        
        self.assertEqual(TimeZoneHelper(datetime.timezone.utc).name, 'UTC')
        self.assertEqual(TimeZoneHelper('Australia/Sydney').name, 'Australia/Sydney')
        self.assertEqual(TimeZoneHelper('US/Eastern').name, 'US/Eastern')
    
    def test_now__utc(self):
        """
        Test the TimeZoneHelper ``now`` method, using the UTC timezone.
        """
        
        now = timezone.now()
        helper_now = TimeZoneHelper(datetime.timezone.utc).now()
        
        self.assertAlmostEqual(helper_now, now, delta=datetime.timedelta(seconds=1))
    
    def test_now__local(self):
        """
        Test the TimeZoneHelper ``now`` method, using a local timezone.
        """
        
        now = timezone.now()
        helper_now = TimeZoneHelper('Australia/Sydney').now()
        local_now = now.astimezone(ZoneInfo('Australia/Sydney'))
        
        self.assertAlmostEqual(helper_now, local_now, delta=datetime.timedelta(seconds=1))
    
    def test_today__utc(self):
        """
        Test the TimeZoneHelper ``today`` method, using the UTC timezone.
        NOTE: This test is subject to some inaccuracy if run at the precise
        moment that caused the two compared datetimes to generate on either
        side of a day interval.
        """
        
        # It would be pretty unlucky for these two datetimes to generate
        # either side of a day interval.
        today = timezone.now().date()
        helper_today = TimeZoneHelper(datetime.timezone.utc).today()
        
        self.assertEqual(helper_today, today)
    
    def test_today__local(self):
        """
        Test the TimeZoneHelper ``today`` method, using a local timezone.
        NOTE: This test is subject to some inaccuracy if run at the precise
        moment that caused the two compared datetimes to generate on either
        side of a day interval.
        """
        
        # It would be pretty unlucky for these two datetimes to generate
        # either side of a day interval.
        now = timezone.now()
        helper_today = TimeZoneHelper('Australia/Sydney').today()
        local_today = now.astimezone(ZoneInfo('Australia/Sydney')).date()
        
        self.assertEqual(helper_today, local_today)
    
    def test_stringify(self):
        """
        Test coercing TimeZoneHelper to a string.
        """
        
        self.assertEqual(str(TimeZoneHelper(datetime.timezone.utc)), 'UTC')
        self.assertEqual(str(TimeZoneHelper('Australia/Sydney')), 'Australia/Sydney')
        self.assertEqual(str(TimeZoneHelper('US/Eastern')), 'US/Eastern')


class SetupTestAppTestCase(SimpleTestCase):
    
    # Use a package *outside* the tests package, since djem's tests themselves
    # make use of setup_test_app()
    package = 'djem.utils'
    
    def test_explicit_label(self):
        """
        Test when an explicit app label is provided. The provided label should
        be used directly.
        """
        
        self.assertNotIn('__test_label__', apps.app_configs)
        setup_test_app(self.package, '__test_label__')
        self.assertIn('__test_label__', apps.app_configs)
        
        # Uninstall the app again to avoid polluting other tests
        apps.app_configs.pop('__test_label__')
    
    def test_explicit_label__duplicate(self):
        """
        Test when called multiple times with an explicit app label provided.
        The second attempt should have no effect.
        """
        
        n = len(apps.app_configs)
        
        setup_test_app(self.package, '__test_label__')
        self.assertEqual(len(apps.app_configs), n + 1)
        
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')  # ensure all warnings are triggered
            
            setup_test_app(self.package, '__test_label__')
            
            self.assertEqual(len(w), 1)
            self.assertTrue(w[0].message, 'Attempted setup of duplicate test app "__test_label__".')
            self.assertTrue(w[0].category, RuntimeWarning)
        
        self.assertEqual(len(apps.app_configs), n + 1)
        
        # Uninstall the app again to avoid polluting other tests
        apps.app_configs.pop('__test_label__')
    
    def test_implicit_label(self):
        """
        Test when no explicit app label is provided. The label of the
        containing app, suffixed with "_tests", should be used.
        """
        
        self.assertNotIn('djem_tests', apps.app_configs)
        setup_test_app(self.package)
        self.assertIn('djem_tests', apps.app_configs)
        
        # Uninstall the app again to avoid polluting other tests
        apps.app_configs.pop('djem_tests')
    
    def test_implicit_label__duplicate(self):
        """
        Test when called multiple times with no explicit app label provided.
        The second attempt should have no effect.
        """
        
        n = len(apps.app_configs)
        
        setup_test_app(self.package)
        self.assertEqual(len(apps.app_configs), n + 1)
        
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')  # ensure all warnings are triggered
            
            setup_test_app(self.package)
            
            self.assertEqual(len(w), 1)
            self.assertTrue(w[0].message, 'Attempted setup of duplicate test app "djem_tests".')
            self.assertTrue(w[0].category, RuntimeWarning)
        
        self.assertEqual(len(apps.app_configs), n + 1)
        
        # Uninstall the app again to avoid polluting other tests
        apps.app_configs.pop('djem_tests')



from django import forms

from djem.utils.dt import TIMEZONE_CHOICES, get_tz_helper


# Based on django-timezone-field
# https://github.com/mfogel/django-timezone-field
# Not used by models.TimeZoneField, but provided for manually defined fields
# in Forms
class TimeZoneField(forms.TypedChoiceField):
    
    def __init__(self, *args, **kwargs):
        
        defaults = {
            'coerce': get_tz_helper,
            'choices': TIMEZONE_CHOICES
        }
        
        defaults.update(kwargs)
        
        super().__init__(*args, **defaults)



import json

from django.contrib import messages
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.http import HttpResponse
from django.template.defaultfilters import mark_safe
from django.test import RequestFactory, TestCase

from djem.ajax import AjaxResponse, ajax_login_required
from djem.utils.tests import MessagingRequestFactory


def _test_view(request):
    
    return HttpResponse()


class AjaxLoginRequiredTestCase(TestCase):
    
    def setUp(self):
        
        self.factory = RequestFactory()
    
    def test_unauthenticated(self):
        """
        Test the ajax_login_required decorator with an unauthenticated user.
        It should return a 403 (forbidden) response.
        """
        
        view = ajax_login_required(_test_view)
        
        request = self.factory.get('/test/')
        request.user = AnonymousUser()
        
        response = view(request)
        
        self.assertEqual(response.status_code, 403)
    
    def test_authenticated(self):
        """
        Test the ajax_login_required decorator with an authenticated user.
        It should return a 200 (success) response.
        """
        
        view = ajax_login_required(_test_view)
        
        request = self.factory.get('/test/')
        request.user = get_user_model().objects.create_user('test1')
        
        response = view(request)
        
        self.assertEqual(response.status_code, 200)


class AjaxResponseTestCase(TestCase):
    
    def setUp(self):
        
        self.factory = MessagingRequestFactory()
    
    def test_response__no_args(self):
        """
        Test that a TypeError is raised when trying to instantiate AjaxResponse
        with no arguments - the "request" argument is required.
        """
        
        def view(r):
            
            return AjaxResponse()
        
        request = self.factory.get('/test/')
        
        with self.assertRaises(TypeError):
            view(request)
    
    def test_response__request_only(self):
        """
        Test that AjaxResponse correctly instantiates when only given the
        "request" argument, and simply contains an empty response body.
        """
        
        def view(r):
            
            return AjaxResponse(r)
        
        request = self.factory.get('/test/')
        response = view(request)
        
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json.loads(response.content.decode()), {})
    
    def test_response__data(self):
        """
        Test that AjaxResponse correctly includes the given data in the response
        body.
        """
        
        def view(r):
            
            return AjaxResponse(r, {'test': 'test'})
        
        request = self.factory.get('/test/')
        response = view(request)
        
        self.assertEqual(response.status_code, 200)
        
        data = json.loads(response.content.decode())
        self.assertEqual(len(data), 1)
        self.assertEqual(data['test'], 'test')
    
    def test_response__bad_data(self):
        """
        Test that a TypeError is raised when trying to instantiate AjaxResponse
        with a non-dict "data" argument.
        """
        
        def view(r):
            
            return AjaxResponse(r, ['test', 'test'])
        
        request = self.factory.get('/test/')
        
        with self.assertRaises(TypeError):
            view(request)
    
    def test_response__success__true(self):
        """
        Test that AjaxResponse correctly includes "success=True" in the response
        body when no other data is given.
        """
        
        def view(r):
            
            return AjaxResponse(r, success=True)
        
        request = self.factory.get('/test/')
        response = view(request)
        
        self.assertEqual(response.status_code, 200)
        
        data = json.loads(response.content.decode())
        self.assertEqual(len(data), 1)
        self.assertIs(data['success'], True)
    
    def test_response__success__false(self):
        """
        Test that AjaxResponse correctly includes the given data in the response
        body.
        """
        
        def view(r):
            
            return AjaxResponse(r, success=False)
        
        request = self.factory.get('/test/')
        response = view(request)
        
        self.assertEqual(response.status_code, 200)
        
        data = json.loads(response.content.decode())
        self.assertEqual(len(data), 1)
        self.assertIs(data['success'], False)
    
    def test_response__success__dumb(self):
        """
        Test that AjaxResponse correctly includes a boolean success flag in the
        response body when the success argument is given as a non-boolean value.
        """
        
        def view(r):
            
            return AjaxResponse(r, success='dumb')
        
        request = self.factory.get('/test/')
        response = view(request)
        
        self.assertEqual(response.status_code, 200)
        
        data = json.loads(response.content.decode())
        self.assertEqual(len(data), 1)
        self.assertIs(data['success'], True)
    
    def test_response__success__data(self):
        """
        Test that AjaxResponse correctly includes a boolean success flag in the
        response body when other data is also provided.
        """
        
        def view(r):
            
            return AjaxResponse(r, {'test': 'test'}, success=True)
        
        request = self.factory.get('/test/')
        response = view(request)
        
        self.assertEqual(response.status_code, 200)
        
        data = json.loads(response.content.decode())
        self.assertEqual(len(data), 2)
        self.assertEqual(data['test'], 'test')
        self.assertIs(data['success'], True)
    
    def test_response__messages(self):
        """
        Test that AjaxResponse correctly includes messages stored by the Django
        messages framework in the response body.
        """
        
        def view(r):
            
            messages.success(request, 'This is a success message.')
            messages.error(request, 'This is an error message.')
            messages.add_message(request, messages.INFO, 'This is an info message.', extra_tags='special')
            
            return AjaxResponse(r)
        
        request = self.factory.get('/test/')
        response = view(request)
        
        self.assertEqual(response.status_code, 200)
        
        data = json.loads(response.content.decode())
        self.assertEqual(len(data), 1)
        self.assertEqual(len(data['messages']), 3)
        
        success = data['messages'][0]
        self.assertEqual(success['message'], 'This is a success message.')
        self.assertEqual(success['tags'], 'success')
        
        error = data['messages'][1]
        self.assertEqual(error['message'], 'This is an error message.')
        self.assertEqual(error['tags'], 'error')
        
        info = data['messages'][2]
        self.assertEqual(info['message'], 'This is an info message.')
        self.assertEqual(info['tags'], 'special info')
    
    def test_response__messages__xss(self):
        """
        Test that AjaxResponse escapes messages added to the response body from
        Django's messages framework.
        """
        
        def view(r):
            
            messages.error(r, 'This is a message <em>with bad HTML</em>.')
            messages.success(r, mark_safe('This is a message <em>with safe HTML</em>.'))
            
            return AjaxResponse(r)
        
        request = self.factory.get('/test/')
        response = view(request)
        
        self.assertEqual(response.status_code, 200)
        
        data = json.loads(response.content.decode())
        self.assertEqual(len(data), 1)
        self.assertEqual(len(data['messages']), 2)
        
        self.assertEqual(data['messages'][0]['message'], 'This is a message &lt;em&gt;with bad HTML&lt;/em&gt;.')
        self.assertEqual(data['messages'][1]['message'], 'This is a message <em>with safe HTML</em>.')
    
    def test_response__messages__data(self):
        """
        Test that AjaxResponse correctly includes messages stored by the Django
        messages framework in the response body when other data is also provided.
        """
        
        def view(r):
            
            messages.success(request, 'This is a success message.')
            
            return AjaxResponse(r, {'test': 'test'})
        
        request = self.factory.get('/test/')
        response = view(request)
        
        self.assertEqual(response.status_code, 200)
        
        data = json.loads(response.content.decode())
        self.assertEqual(len(data), 2)
        self.assertEqual(len(data['messages']), 1)
        self.assertEqual(data['test'], 'test')
        
        success = data['messages'][0]
        self.assertEqual(success['message'], 'This is a success message.')
        self.assertEqual(success['tags'], 'success')



from django.contrib import messages
from django.contrib.messages import constants
from django.http import HttpRequest, HttpResponse
from django.test import Client, TestCase, override_settings
from django.urls import path

from djem.middleware import MemoryStorage


def add_message_view(request):
    
    msg = request.GET['msg']
    messages.info(request, msg)
    
    if request.headers.get('x-requested-with') == 'XMLHttpRequest':
        prefix = 'AJAX'
    else:
        prefix = 'STANDARD'
    
    return HttpResponse(f'{prefix}: no messages')


def add_read_message_view(request):
    
    msg = request.GET['msg']
    messages.info(request, msg)
    
    content = ', '.join([msg.message for msg in messages.get_messages(request)])
    
    if request.headers.get('x-requested-with') == 'XMLHttpRequest':
        prefix = 'AJAX'
    else:
        prefix = 'STANDARD'
    
    return HttpResponse(f'{prefix}: {content}')


urlpatterns = [
    path('messages/add/', add_message_view),
    path('messages/add/read/', add_read_message_view),
]


djem_middleware_settings = override_settings(
    ROOT_URLCONF='djem.tests.test_middleware',
    MIDDLEWARE=[
        'django.contrib.sessions.middleware.SessionMiddleware',
        'djem.middleware.MessageMiddleware'
    ]
)

django_middleware_settings = override_settings(
    ROOT_URLCONF='djem.tests.test_middleware',
    MIDDLEWARE=[
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django.contrib.messages.middleware.MessageMiddleware'
    ]
)


class MemoryStorageTestCase(TestCase):
    
    def test_add(self):
        """
        Test messages can be added and retrieved as expected.
        """
        
        # Create a message store with a fake request instance
        messages = MemoryStorage(HttpRequest())
        
        self.assertEqual(len(messages), 0)
        
        messages.add(constants.INFO, 'Test message')
        self.assertEqual(len(messages), 1)
        
        messages.add(constants.INFO, 'Another test message')
        self.assertEqual(len(messages), 2)
        
        # Read the message store
        message_list = list(messages)
        
        messages.add(constants.INFO, 'A third test message')
        
        self.assertEqual(len(message_list), 2)
        self.assertEqual(len(messages), 3)


class MessageMiddlewareTestCase(TestCase):
    
    def setUp(self):
        
        self.client = Client()
    
    @djem_middleware_settings
    def test_standard_request(self):
        """
        Test that messages added via the message framework, on a standard
        request, can be read back when using djem's MessageMiddleware.
        """
        
        response = self.client.get(
            '/messages/add/read/',
            {'msg': 'test message'}
        )
        
        self.assertEqual(response.content, b'STANDARD: test message')
    
    @djem_middleware_settings
    def test_ajax_request(self):
        """
        Test that messages added via the message framework, on an AJAX
        request, can be read back when using djem's MessageMiddleware.
        """
        
        response = self.client.get(
            '/messages/add/read/',
            {'msg': 'test message'},
            HTTP_X_REQUESTED_WITH='XMLHttpRequest'
        )
        
        self.assertEqual(response.content, b'AJAX: test message')
    
    @djem_middleware_settings
    def test_mixed_requests(self):
        """
        Test that messages added on standard requests and AJAX requests use
        different stores and do not interfere with each other when using djem's
        MessageMiddleware.
        """
        
        # Start with a standard request that adds a message, but doesn't read
        # back the message store
        response = self.client.get(
            '/messages/add/',
            {'msg': 'first standard message'}
        )
        
        self.assertEqual(response.content, b'STANDARD: no messages')
        
        # Next, trigger an AJAX request that adds a message, but also doesn't
        # read back the message store. This message should be lost once the
        # request is completed.
        response = self.client.get(
            '/messages/add/',
            {'msg': 'lost ajax message'},
            HTTP_X_REQUESTED_WITH='XMLHttpRequest'
        )
        
        self.assertEqual(response.content, b'AJAX: no messages')
        
        # Then trigger an AJAX request that adds a message and does read back
        # the message store - it should only see the message it added, not
        # either of the two previous messages.
        response = self.client.get(
            '/messages/add/read/',
            {'msg': 'ajax message'},
            HTTP_X_REQUESTED_WITH='XMLHttpRequest'
        )
        
        self.assertEqual(response.content, b'AJAX: ajax message')
        
        # Finally, trigger another standard request that adds a message and
        # reads back the message store - it should see the two messages added
        # as part of standard requests, and not those added in the AJAX request
        response = self.client.get(
            '/messages/add/read/',
            {'msg': 'second standard message'}
        )
        
        self.assertEqual(response.content, b'STANDARD: first standard message, second standard message')
    
    @django_middleware_settings
    def test_django_mixed_requests(self):
        """
        Test that messages added on standard requests and AJAX requests DO
        interfere with each other when using Django's MessageMiddleware.
        """
        
        # Start with a standard request that adds a message, but doesn't read
        # back the message store
        response = self.client.get(
            '/messages/add/',
            {'msg': 'first standard message'}
        )
        
        self.assertEqual(response.content, b'STANDARD: no messages')
        
        # Then trigger an AJAX request that adds a message and does read back
        # the message store - it sees all messages
        response = self.client.get(
            '/messages/add/read/',
            {'msg': 'ajax message'},
            HTTP_X_REQUESTED_WITH='XMLHttpRequest'
        )
        
        self.assertEqual(response.content, b'AJAX: first standard message, ajax message')
        
        # Finally, trigger another standard request that adds a message and
        # reads back the message store - it only sees its own message, since
        # the others were consumed by the intervening AJAX request
        response = self.client.get(
            '/messages/add/read/',
            {'msg': 'second standard message'}
        )
        
        self.assertEqual(response.content, b'STANDARD: second standard message')



from functools import wraps

from django.contrib.messages import get_messages
from django.http import HttpResponseForbidden, JsonResponse
from django.template.defaultfilters import conditional_escape


def ajax_login_required(view_fn):
    """
    Simple decorator that returns ``HttpResponseForbidden`` if the user is not
    authenticated, and executes the decorated function if they are. Designed
    for use by views accessed with AJAX requests, where redirecting to the
    login view is not useful.
    """
    
    @wraps(view_fn)
    def wrapper(request, *args, **kwargs):
        
        if not request.user.is_authenticated:
            return HttpResponseForbidden()
        
        return view_fn(request, *args, **kwargs)
    
    return wrapper


class AjaxResponse(JsonResponse):
    """
    Extension of Django's ``JsonResponse`` with the following additional features:
    
    The first positional argument should be a Django ``HttpRequest`` instance.
    
    The ``data`` argument is optional. It may be modified (see below) before
    being passed to the parent ``JsonResponse``. If provided, it must always be
    a ``dict`` instance - using the ``safe`` argument of ``JsonResponse`` to
    JSON-encode other types is not supported.
    
    The optional argument ``success`` can be set to add a "success" attribute
    to the ``data`` dictionary.
    
    A "messages" attribute is automatically added to the ``data`` dictionary if
    the Django message framework store contains any messages at the time of the
    response. The messages in the store will be added as a list of dictionaries
    with the keys "message" and "tags".
    
    With the exception of ``safe``, as noted above, it supports all arguments of
    ``JsonResponse``.
    """
    
    def __init__(self, request, data=None, success=None, **kwargs):
        
        if data is None:
            data = {}
        
        if not isinstance(data, dict):
            msg = 'Non-dict data types are not supported'
            raise TypeError(msg)
        
        if success is not None:
            data['success'] = bool(success)
        
        messages = get_messages(request)
        if messages:
            # "Serialise" messages. Escape the message string as it would be
            # if the messages were rendered by an auto-escaping Django template.
            data['messages'] = [{'message': conditional_escape(m.message), 'tags': m.tags} for m in messages]
        
        super().__init__(data, **kwargs)



from django.conf import settings
from django.contrib.messages.storage import default_storage
from django.contrib.messages.storage.base import BaseStorage


class MemoryStorage(BaseStorage):
    """
    A messages framework storage backend that does not implement any persistent
    storage of the messages added to it. Messages are stored in memory only -
    if they are not read within the same request, they are lost.
    """
    
    def _get(self, *args, **kwargs):
        
        return None, True
    
    def _store(self, messages, response, *args, **kwargs):
        
        return []


class MessageMiddleware:
    """
    Middleware that handles temporary messages, differentiating between those
    added as part of an AJAX request vs those that are not. Different storage
    backends are used for each.
    """
    
    def __init__(self, get_response):
        
        self.get_response = get_response
    
    def __call__(self, request):
        
        is_ajax = request.headers.get('x-requested-with') == 'XMLHttpRequest'
        
        if is_ajax:
            request._messages = MemoryStorage(request)
        else:
            request._messages = default_storage(request)
        
        response = self.get_response(request)
        
        #
        # The following is copied verbatim from Django's own MessageMiddleware
        # (apart from the pragma)
        #
        
        # A higher middleware layer may return a request which does not contain
        # messages storage, so make no assumption that it will be there.
        if hasattr(request, '_messages'):  # pragma: no cover
            unstored_messages = request._messages.update(response)
            if unstored_messages and settings.DEBUG:
                msg = 'Not all temporary messages could be stored.'
                raise ValueError(msg)
        
        return response



import operator
import shutil
import textwrap

from django.utils.encoding import force_str

from djem.utils.misc import UNDEFINED


class FullWidth:
    """
    Max width constant indicating the full width of the terminal.
    """
    
    pass


class BlankRow:
    """
    Row constant representing a blank row.
    """
    
    pass


class HorizontalRule:
    """
    Row constant representing a horizontal rule.
    """
    
    pass


class RowWrapper:
    """
    Helper class for formatting full-width rows such as titles, descriptions and
    footers.
    """
    
    ALIGN_MAP = {
        'left': 'ljust',
        'right': 'rjust',
        'center': 'center',
        'centre': 'center'
    }
    
    def __init__(self, value, alignment='left'):
        
        self.raw_value = value
        self.set_alignment(alignment)
    
    def set_alignment(self, alignment):
        
        try:
            self._align = self.ALIGN_MAP[alignment]
        except KeyError:
            self._align = 'ljust'
    
    def get_rows(self, table_width):
        
        rows = []
        
        # Subtract 4 from the table width to account for the "| " and " |"
        table_width -= 4
        
        # Split value into multiple rows based on line breaks. Wrap any of the
        # rows that are longer than the shortened table width, and align them
        # within that width.
        for line in self.raw_value.split('\n'):
            for sub_line in textwrap.wrap(line, table_width):
                aligned = getattr(sub_line, self._align)(table_width)
                rows.append(f'| {aligned} |')
        
        return rows


class Table:
    
    FULL_WIDTH = FullWidth()
    BR = BlankRow()
    HR = HorizontalRule()
    
    MIN_COLUMN_WIDTH = 4  # allows for a single character and "..." to indicate truncation
    
    def __init__(self, headings=None, title=None, footer=None, max_width=FULL_WIDTH):
        
        self._rows = []
        self._columns = []
        
        if headings:
            self.add_headings(headings)
        else:
            self._headings = None
        
        if title:
            self.set_title(title)
        else:
            self.title = None
        
        if footer:
            self.set_footer(footer)
        else:
            self.footer = None
        
        self._raw_max_width = max_width
    
    def _update_col_metadata(self, col_data, value=UNDEFINED, heading=UNDEFINED):
        
        if value is not UNDEFINED:
            width = len(force_str(value))
        elif heading is not UNDEFINED:
            heading = force_str(heading)
            width = 0
            
            col_data['heading'] = heading
        else:
            msg = 'Invalid arguments provided.'
            raise TypeError(msg)
        
        if width > col_data.get('raw_width', 0):
            col_data['raw_width'] = width
        
        return col_data
    
    def _reduce_column_widths(self, reduction, width_map):
        
        # Reduce the widths of each column until the overall table width no
        # longer exceeds the maximum. Target the widest columns first. When
        # multiple columns share the same width, target those that come later
        # in the table.
        sorted_widths = sorted(width_map.keys(), reverse=True)
        
        for i, col_width in enumerate(sorted_widths):
            group_cols = width_map[col_width]
            
            try:
                next_width = sorted_widths[i + 1]
            except IndexError:
                next_width = None
                diff = 0
                group_diff = reduction
            else:
                diff = abs(col_width - next_width)
                group_diff = diff * len(group_cols)
            
            if group_diff < reduction:
                # Reduce the widths of all columns in the group to match the
                # width of the next-widest column
                col_reduction = diff
                remainder = 0
                reduction -= group_diff
                
                # Add these columns to the next group, as they will all have
                # the same width
                width_map[next_width].extend(group_cols)
            else:
                # What's left of the reduction amount cannot be divided evenly
                # amongst all columns in the group - spread it out as evenly
                # as possible
                col_reduction, remainder = divmod(reduction, len(group_cols))
                reduction = 0
            
            if col_reduction:
                for col in group_cols:
                    col['render_width'] -= col_reduction
            
            if remainder:
                # Sort the list of columns by descending position, so the
                # remainder comes off later columns before earlier ones
                sorted_cols = sorted(group_cols, key=operator.itemgetter('pos'), reverse=True)
                for col in sorted_cols[:remainder]:
                    col['render_width'] -= 1
            
            if not reduction:
                break
    
    def set_title(self, title, alignment='centre'):
        
        self.title = RowWrapper(title, alignment)
    
    def set_footer(self, footer, alignment='right'):
        
        self.footer = RowWrapper(footer, alignment)
    
    def add_headings(self, headings):
        
        cols = self._columns
        
        if not cols:
            self._columns = [self._update_col_metadata({}, heading=h) for h in headings]
        elif len(cols) != len(headings):
            msg = 'Number of headings does not match previously given number of columns.'
            raise ValueError(msg)
        else:
            for i, heading in enumerate(headings):
                self._update_col_metadata(cols[i], heading=heading)
        
        self._headings = headings
    
    def add_row(self, row):
        
        if row is self.BR or row is self.HR:
            self._rows.append(row)
            return
        
        if self._columns and len(self._columns) != len(row):
            msg = 'Number of columns in row does not match previously given rows.'
            raise ValueError(msg)
        
        for i, value in enumerate(row):
            try:
                col_data = self._columns[i]
            except IndexError:
                self._columns.append({})
                col_data = self._columns[i]
            
            self._update_col_metadata(col_data, value=value)
        
        self._rows.append(row)
    
    def add_full_width_row(self, value, alignment='left'):
        
        row = RowWrapper(value, alignment)
        
        self._rows.append(row)
    
    def add_rows(self, rows):
        
        for row in rows:
            self.add_row(row)
    
    def get_rows(self):
        
        rows = [self.HR]
        
        if self._headings:
            rows.extend((self._headings, self.HR))
        
        rows.extend(self._rows)
        rows.append(self.HR)
        
        return rows
    
    def get_max_widths(self):
        
        outer_max_width = self._raw_max_width
        
        if outer_max_width is self.FULL_WIDTH:
            term_columns, _ = shutil.get_terminal_size()
            outer_max_width = int(term_columns)
        
        # The max width of actual data in the table is the outer max less the
        # space required for formatting the table itself
        
        # +4 for the "| " and " |" at the start and end of the line, and +3 for
        # the " | " spacer between each column
        formatting_buffer = 4 + ((len(self._columns) - 1) * 3)
        
        max_width = outer_max_width - formatting_buffer
        
        return outer_max_width, max_width
    
    def calculate_render_widths(self):
        
        total_width = 0
        min_width = 0
        width_map = {}
        
        # Determine full render widths per column as the max of the raw data
        # width and the heading width
        for i, col in enumerate(self._columns):
            col['pos'] = i
            col_width = max(col.get('raw_width', 0), len(col.get('heading', '')))
            col['render_width'] = col_width
            total_width += col_width
            min_width += min(col_width, self.MIN_COLUMN_WIDTH)
            
            # Map column widths to the columns with that width
            try:
                width_map[col_width].append(col)
            except KeyError:
                width_map[col_width] = [col]
        
        _, max_width = self.get_max_widths()
        
        if max_width < min_width:
            msg = 'Minimum table width exceeds maximum: table cannot be drawn.'
            raise ValueError(msg)
        
        # Get the amount by which the total table width needs to be reduced to
        # fit within the max width
        reduction = total_width - max_width
        
        if reduction <= 0:
            # No reduction of widths necessary
            return total_width
        
        total_width -= reduction
        
        self._reduce_column_widths(reduction, width_map)
        
        return total_width
    
    def build_table(self):
        
        formatting_buffer = 4 + ((len(self._columns) - 1) * 3)
        
        total_raw_width = self.calculate_render_widths()
        table_width = total_raw_width + formatting_buffer
        
        br = f'|{" " * (table_width - 2)}|'
        hr = f'+{"-" * (table_width - 2)}+'
        
        output = []
        
        if self.title:
            output.append(hr)
            output.extend(self.title.get_rows(table_width))
        
        for row in self.get_rows():
            if row is self.BR:
                output.append(br)
            elif row is self.HR:
                output.append(hr)
            elif isinstance(row, RowWrapper):
                output.extend(row.get_rows(table_width))
            else:
                row_str = []
                for i, col in enumerate(self._columns):
                    value = force_str(row[i]).replace('\n', '\\n').replace('\r', '\\r')
                    width = col['render_width']
                    
                    if len(value) > width:
                        value = f'{value[:width - 3]}...'
                    
                    row_str.append(value.ljust(width))
                
                output.append(f'| {" | ".join(row_str)} |')
        
        if self.footer:
            output.extend(self.footer.get_rows(table_width))
            output.append(hr)
        
        return '\n'.join(output)
